# This file is part of ReText
# Copyright: 2014 Lukas Holecek
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from FakeVim import FAKEVIM_PYQT_VERSION, FakeVimHandler, FakeVimProxy, MessageError

if FAKEVIM_PYQT_VERSION != 6:
    raise ImportError("FakeVim must be compiled with Qt 6")

from PyQt6.QtCore import QDir, QObject, QRegExp, Qt
from PyQt6.QtGui import QPainter, QPen, QTextCursor
from PyQt6.QtWidgets import (
    QLabel,
    QLineEdit,
    QMessageBox,
    QStatusBar,
    QTextEdit,
    QWidget,
)


class FakeVimMode:
    @staticmethod
    def init(window):
        window.setStatusBar(StatusBar())

    @staticmethod
    def exit(window):
        window.statusBar().deleteLater()

class Proxy (FakeVimProxy):
    """ Used by FakeVim to modify or retrieve editor state. """
    def __init__(self, window, editor, handler):
        super().__init__(handler.handler())
        self.__handler = handler
        self.__window = window
        self.__editor = editor

        self.__statusMessage = ""
        self.__statusData = ""
        self.__cursorPosition =  -1
        self.__cursorAnchor =  -1
        self.__eventFilter = None

        self.__lastSavePath = ""

    def showMessage(self, messageLevel, message):
        self.__handler.handler().showMessage(messageLevel, message)

    def needSave(self):
        return self.__editor.document().isModified()

    def maybeCloseEditor(self):
        if self.needSave():
            self.showMessage( MessageError,
                    self.tr("No write since last change (add ! to override)") )
            self.__updateStatusBar()

            return False

        return True

    def commandQuit(self):
        self.__handler.quit()

    def commandWrite(self):
        self.__handler.save()
        return not self.needSave()

    def handleExCommand(self, cmd):
        if cmd.matches("q", "quit"):
            if cmd.hasBang or self.maybeCloseEditor():
                self.commandQuit()
        elif cmd.matches("w", "write"):
            self.commandWrite()
        elif cmd.cmd == "wq":
            self.commandWrite() and self.commandQuit()
        else:
            return False
        return True

    def enableBlockSelection(self, cursor):
        self.__handler.setBlockSelection(True)
        self.__editor.setTextCursor(cursor)

    def disableBlockSelection(self):
        self.__handler.setBlockSelection(False)

    def blockSelection(self):
        self.__handler.setBlockSelection(True)
        return self.__editor.textCursor()

    def hasBlockSelection(self):
        return self.__handler.hasBlockSelection()

    def commandBufferChanged(self, msg, cursorPosition, cursorAnchor, messageLevel, eventFilter):
        # Give focus back to editor if closing command line.
        if self.__cursorPosition != -1 and cursorPosition == -1:
            self.__editor.setFocus()

        self.__cursorPosition = cursorPosition
        self.__cursorAnchor = cursorAnchor
        self.__statusMessage = msg
        self.__updateStatusBar()
        self.__eventFilter = eventFilter

    def statusDataChanged(self, msg):
        self.__statusData = msg
        self.__updateStatusBar()

    def extraInformationChanged(self, msg):
        QMessageBox.information(self.__window, self.tr("Information"), msg)

    def highlightMatches(self, pattern):
        self.__handler.highlightMatches(pattern)

    def __updateStatusBar(self):
        self.__window.statusBar().setStatus(
                self.__statusMessage, self.__statusData,
                self.__cursorPosition, self.__cursorAnchor, self.__eventFilter)

class BlockSelection (QWidget):
    def __init__(self, editor):
        super().__init__(editor.viewport())
        self.__editor = editor
        self.__lineWidth = 4

    def updateSelection(self, tc):
        # block selection rectangle
        rect = self.__editor.cursorRect(tc)
        w = rect.width()
        tc2 = QTextCursor(tc)
        tc2.setPosition(tc.anchor())
        rect = rect.united( self.__editor.cursorRect(tc2) )
        x = self.__lineWidth / 2
        rect.adjust(-x, -x, x - w, x)

        self.setGeometry(rect)

    def paintEvent(self, paintEvent):
        painter = QPainter(self)
        painter.setClipRect(paintEvent.rect())

        color = self.__editor.palette().text()
        painter.setPen(QPen(color, self.__lineWidth))
        painter.drawRect(self.rect())

class ReTextFakeVimHandler (QObject):
    """ Editor widget driven by FakeVim. """
    def __init__(self, editor, window):
        super().__init__(window)

        self.__window = window
        self.__editor = editor

        self.__blockSelection = BlockSelection(self.__editor)
        self.__blockSelection.hide()

        self.__searchSelections = []

        fm = self.__editor.fontMetrics()
        self.__cursorWidth = fm.averageCharWidth()
        self.__oldCursorWidth = self.__editor.cursorWidth()
        self.__editor.setCursorWidth(self.__cursorWidth)

        self.__handler = FakeVimHandler(self.__editor, self)
        self.__proxy = Proxy(self.__window, self.__editor, self)

        self.__handler.installEventFilter()
        self.__handler.setupWidget()
        self.__handler.handleCommand(
                f'source {QDir.homePath()}/.vimrc')

        self.__saveAction = None
        self.__quitAction = None

        # Update selections if cursor changes because of current line can be highlighted.
        self.__editor.cursorPositionChanged.connect(self.__updateSelections)

    def remove(self):
        self.__editor.setOverwriteMode(False)
        self.__editor.setCursorWidth(self.__oldCursorWidth)
        self.__blockSelection.deleteLater()
        self.__updateSelections([])
        self.deleteLater()

    def handler(self):
        return self.__handler

    def setBlockSelection(self, enabled):
        self.__editor.setCursorWidth(self.__cursorWidth)
        self.__blockSelection.setVisible(enabled)

        if enabled:
            self.__blockSelection.updateSelection(self.__editor.textCursor())

            # Shift text cursor into the block selection.
            tc = self.__editor.textCursor()
            if self.__columnForPosition(tc.anchor()) < self.__columnForPosition(tc.position()):
                self.__editor.setCursorWidth(-self.__cursorWidth)

    def setSaveAction(self, saveAction):
        self.__saveAction = saveAction

    def setQuitAction(self, quitAction):
        self.__quitAction = quitAction

    def save(self):
        if self.__saveAction:
            self.__saveAction.trigger()

    def quit(self):
        if self.__quitAction:
            self.__quitAction.trigger()

    def hasBlockSelection(self):
        return self.__blockSelection.isVisible()

    def highlightMatches(self, pattern):
        cur = self.__editor.textCursor()

        re = QRegExp(pattern)
        cur = self.__editor.document().find(re)
        a = cur.position()

        searchSelections = []

        while not cur.isNull():
            if cur.hasSelection():
                selection = QTextEdit.ExtraSelection()
                selection.format.setBackground(Qt.GlobalColor.yellow)
                selection.format.setForeground(Qt.GlobalColor.black)
                selection.cursor = cur
                searchSelections.append(selection)
            else:
                cur.movePosition(QTextCursor.MoveOperation.NextCharacter)

            cur = self.__editor.document().find(re, cur)
            b = cur.position()

            if a == b:
                cur.movePosition(QTextCursor.MoveOperation.NextCharacter)
                cur = self.__editor.document().find(re, cur)
                b = cur.position()

                if (a == b):
                    break
            a = b

        self.__updateSelections(searchSelections)

    def __updateSelections(self, searchSelections = None):
        oldSelections = self.__editor.extraSelections()

        for selection in self.__searchSelections:
            for i in range(len(oldSelections) - 1, 0, -1):
                if selection.cursor == oldSelections[i].cursor:
                    oldSelections.pop(i)
                    break

        if searchSelections is not None:
            self.__searchSelections = searchSelections

        self.__editor.setExtraSelections(oldSelections + self.__searchSelections)

    def __columnForPosition(self, position):
        return position - self.__editor.document().findBlock(position).position()

class StatusBar (QStatusBar):
    def __init__(self):
        super().__init__()

        self.__statusMessageLabel = QLabel(self)
        self.__statusDataLabel = QLabel(self)
        self.__commandLine = QLineEdit(self)

        self.addPermanentWidget(self.__statusMessageLabel, 1)
        self.addPermanentWidget(self.__commandLine, 1)
        self.addPermanentWidget(self.__statusDataLabel)

        self.__commandLine.hide()

    def setStatus(self, statusMessage, statusData, cursorPosition, cursorAnchor, eventFilter):
        commandMode = cursorPosition != -1
        self.__commandLine.setVisible(commandMode)
        self.__statusMessageLabel.setVisible(not commandMode)

        if commandMode:
            self.__commandLine.installEventFilter(eventFilter)
            self.__commandLine.setFocus()
            self.__commandLine.setText(statusMessage)
            self.__commandLine.setSelection(cursorPosition, cursorAnchor - cursorPosition)
        else:
            self.__statusMessageLabel.setText(statusMessage)

        self.__statusDataLabel.setText(statusData)



# This file is part of ReText
# Copyright: 2016-2023 Maurice van der Pot
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from PyQt6.QtCore import QPoint


class SyncScroll:

    def __init__(self, previewFrame,
                       editorPositionToSourceLineFunc,
                       sourceLineToEditorPositionFunc):
        self.posmap = {}
        self.frame = previewFrame
        self.editorPositionToSourceLine = editorPositionToSourceLineFunc
        self.sourceLineToEditorPosition = sourceLineToEditorPositionFunc

        self.previewPositionBeforeLoad = QPoint()
        self.contentIsLoading = False

        self.editorViewportHeight = 0
        self.editorViewportOffset = 0
        self.editorCursorPosition = 0

        self.frame.contentsSizeChanged.connect(self._handlePreviewResized)
        self.frame.loadStarted.connect(self._handleLoadStarted)
        self.frame.loadFinished.connect(self._handleLoadFinished)

    def isActive(self):
        return bool(self.posmap)

    def handleEditorResized(self, editorViewportHeight):
        self.editorViewportHeight = editorViewportHeight
        self._updatePreviewScrollPosition()

    def handleEditorScrolled(self, editorViewportOffset):
        self.editorViewportOffset = editorViewportOffset
        return self._updatePreviewScrollPosition()

    def handleCursorPositionChanged(self, editorCursorPosition):
        self.editorCursorPosition = editorCursorPosition
        return self._updatePreviewScrollPosition()

    def _handleLoadStarted(self):
        # Store the current scroll position so it can be restored when the new
        # content is presented
        self.previewPositionBeforeLoad = self.frame.scrollPosition()
        self.contentIsLoading = True

    def _handleLoadFinished(self):
        self.frame.setScrollPosition(self.previewPositionBeforeLoad)
        self.contentIsLoading = False
        self.frame.getPositionMap(self._setPositionMap)

    def _handlePreviewResized(self):
        self.frame.getPositionMap(self._setPositionMap)
        self._updatePreviewScrollPosition()
        if not self.posmap and self.frame.scrollPosition().y() == 0:
            self.frame.setScrollPosition(self.previewPositionBeforeLoad)

    def _linearScale(self, fromValue, fromMin, fromMax, toMin, toMax):
        fromRange = fromMax - fromMin
        toRange = toMax - toMin

        toValue = toMin

        if fromRange:
            toValue += ((fromValue - fromMin) * toRange) / float(fromRange)

        return toValue

    def _updatePreviewScrollPosition(self):
        if not self.posmap:
            # Loading new content resets the scroll position to the top. If we
            # don't have a posmap to calculate the new best position, then
            # restore the position stored at the beginning of the load.
            if self.contentIsLoading:
                self.frame.setScrollPosition(self.previewPositionBeforeLoad)
            return

        textedit_pixel_to_scroll_to = self.editorCursorPosition

        if textedit_pixel_to_scroll_to < self.editorViewportOffset:
            textedit_pixel_to_scroll_to = self.editorViewportOffset

        last_viewport_pixel = self.editorViewportOffset + self.editorViewportHeight
        if textedit_pixel_to_scroll_to > last_viewport_pixel:
            textedit_pixel_to_scroll_to = last_viewport_pixel

        line_to_scroll_to = self.editorPositionToSourceLine(textedit_pixel_to_scroll_to)

        # Do a binary search through the posmap to find the nearest line above
        # and below the line to scroll to for which the rendered position is
        # known.
        posmap_lines = [0] + sorted(self.posmap.keys())
        min_index = 0
        max_index = len(posmap_lines) - 1
        while max_index - min_index > 1:
            current_index = int((min_index + max_index) / 2)
            if posmap_lines[current_index] > line_to_scroll_to:
                max_index = current_index
            else:
                min_index = current_index

        # number of nearest line above and below for which we have a position
        min_line = posmap_lines[min_index]
        max_line = posmap_lines[max_index]

        min_textedit_pos = self.sourceLineToEditorPosition(min_line)
        max_textedit_pos = self.sourceLineToEditorPosition(max_line)

        # rendered pixel position of nearest line above and below
        min_preview_pos = self.posmap[min_line]
        max_preview_pos = self.posmap[max_line]

        # calculate rendered pixel position of line corresponding to cursor
        # (0 == top of document)
        preview_pixel_to_scroll_to = self._linearScale(textedit_pixel_to_scroll_to,
                                                      min_textedit_pos, max_textedit_pos,
                                                      min_preview_pos, max_preview_pos)

        distance_to_top_of_viewport_editor = textedit_pixel_to_scroll_to - self.editorViewportOffset
        distance_to_top_of_viewport_preview = distance_to_top_of_viewport_editor / self.frame.zoomFactor()
        preview_scroll_offset = preview_pixel_to_scroll_to - distance_to_top_of_viewport_preview

        pos = self.frame.scrollPosition()
        pos.setY(preview_scroll_offset)
        self.frame.setScrollPosition(pos)

    def _setPositionMap(self, posmap):
        self.posmap = posmap
        if posmap:
            self.posmap[0] = 0



# This file is part of ReText
# Copyright: 2012-2025 Dmitry Shachnev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from PyQt6.QtWidgets import (
    QCheckBox,
    QDialog,
    QDialogButtonBox,
    QLabel,
    QLineEdit,
    QTextEdit,
    QVBoxLayout,
)

from ReText import globalSettings
from ReText.highlighter import ReTextHighlighter


class HtmlDialog(QDialog):
    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        self.resize(700, 600)
        verticalLayout = QVBoxLayout(self)
        self.textEdit = QTextEdit(self)
        self.textEdit.setReadOnly(True)
        self.textEdit.setFont(globalSettings.getEditorFont())
        self.hl = ReTextHighlighter(self.textEdit.document())
        self.hl.docType = 'html'
        verticalLayout.addWidget(self.textEdit)
        buttonBox = QDialogButtonBox(self)
        buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Close)
        buttonBox.rejected.connect(self.close)
        verticalLayout.addWidget(buttonBox)

class LocaleDialog(QDialog):
    def __init__(self, parent, defaultText=None):
        QDialog.__init__(self, parent)
        verticalLayout = QVBoxLayout(self)
        labelText = self.tr('Enter locale name (example: en_US)') + '\n'
        labelText += self.tr('It is possible to specify multiple languages, separated by comma.')
        verticalLayout.addWidget(QLabel(labelText, self))
        self.localeEdit = QLineEdit(self)
        if defaultText:
            self.localeEdit.setText(defaultText)
        verticalLayout.addWidget(self.localeEdit)
        self.checkBox = QCheckBox(self.tr('Set as default'), self)
        verticalLayout.addWidget(self.checkBox)
        buttonBox = QDialogButtonBox(self)
        buttonBox.setStandardButtons(
            QDialogButtonBox.StandardButton.Cancel | QDialogButtonBox.StandardButton.Ok
        )
        verticalLayout.addWidget(buttonBox)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)


class EncodingDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        verticalLayout = QVBoxLayout(self)
        verticalLayout.addWidget(QLabel(self.tr('Enter encoding name:'), self))
        self.encodingEdit = QLineEdit(self)
        self.encodingEdit.textChanged.connect(self.handleTextChanged)
        verticalLayout.addWidget(self.encodingEdit)
        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel
                                          | QDialogButtonBox.StandardButton.Ok)
        verticalLayout.addWidget(self.buttonBox)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

    def handleTextChanged(self, value):
        button = self.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        try:
            "1".encode(value)
        except LookupError:
            button.setEnabled(False)
        else:
            button.setEnabled(True)



# vim: ts=4:sw=4:expandtab
#
# This file is part of ReText
# Copyright: 2012-2025 Dmitry Shachnev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import weakref

from markups import (
    AsciiDocMarkup,
    MarkdownMarkup,
    ReStructuredTextMarkup,
    TextileMarkup,
)
from PyQt6.QtCore import (
    QFileInfo,
    QMimeDatabase,
    QPoint,
    QPointF,
    QRect,
    QSize,
    Qt,
    pyqtSignal,
)
from PyQt6.QtGui import (
    QAction,
    QColor,
    QGuiApplication,
    QImage,
    QKeyEvent,
    QMouseEvent,
    QPainter,
    QPalette,
    QTextCursor,
    QTextFormat,
    QTextOption,
    QWheelEvent,
)
from PyQt6.QtWidgets import QApplication, QFileDialog, QLabel, QTextEdit, QWidget

from ReText import globalSettings, settings, tablemode

try:
    from ReText.fakevimeditor import ReTextFakeVimHandler
except ImportError:
    ReTextFakeVimHandler = None

colors = {
    # Editor
    'marginLine':           {'light': '#dcd2dc', 'dark': '#3daee9'},
    'currentLineHighlight': {'light': '#ffffc8', 'dark': '#31363b'},
    'infoArea':             {'light': '#aaaaff55', 'dark': '#aa557f2a'},
    'statsArea':            {'light': '#aaffaa55', 'dark': '#aa7f552a'},
    'lineNumberArea':       {'light': '#00ffff', 'dark': '#31363b'},
    'lineNumberAreaText':   {'light': '#008080', 'dark': '#bdc3c7'},
    # Highlighter
    'htmlTags':             {'light': '#800080', 'dark': '#d070d0'},
    'htmlSymbols':          {'light': '#008080', 'dark': '#70d0a0'},
    'htmlStrings':          {'light': '#808000', 'dark': '#d0d070'},
    'htmlComments':         {'light': '#a0a0a4', 'dark': '#b0b0aa'},
    'codeSpans':            {'light': '#505050', 'dark': '#afafaf'},
    'markdownHeaders':      {'light': '#000000', 'dark': '#ffffff'},
    'markdownLinks':        {'light': '#000090', 'dark': '#8080ff'},
    'blockquotes':          {'light': '#808080', 'dark': '#b0b0b0'},
    'restDirectives':       {'light': '#800080', 'dark': '#d070d0'},
    'restRoles':            {'light': '#800000', 'dark': '#d07070'},
    'whitespaceOnEnd':      {'light': '#80e1e1a5', 'dark': '#8096966e'},
    # Preview
    'urlPopupArea':         {'light': '#fafafafa', 'dark': '#fa323232'},
    'urlPopupBorder':       {'light': '#64323232', 'dark': '#64fafafa'},
}

colorValues = {}

def updateColorScheme(settings=settings):
    palette = QApplication.palette()
    windowColor = palette.color(QPalette.ColorRole.Window)
    themeVariant = 'light' if windowColor.lightness() > 150 else 'dark'
    settings.beginGroup('ColorScheme')
    for key in colors:
        if settings.contains(key):
            colorValues[key] = settings.value(key, type=QColor)
        else:
            colorValues[key] = QColor(colors[key][themeVariant])
    settings.endGroup()

def getColor(colorName):
    if not colorValues:
        updateColorScheme()
    return colorValues[colorName]

def documentIndentMore(document, cursor, globalSettings=globalSettings):
    if cursor.hasSelection():
        block = document.findBlock(cursor.selectionStart())
        end = document.findBlock(cursor.selectionEnd()).next()
        cursor.beginEditBlock()
        while block != end:
            cursor.setPosition(block.position())
            if globalSettings.tabInsertsSpaces:
                cursor.insertText(' ' * globalSettings.tabWidth)
            else:
                cursor.insertText('\t')
            block = block.next()
        cursor.endEditBlock()
    else:
        indent = globalSettings.tabWidth - (cursor.positionInBlock()
            % globalSettings.tabWidth)
        if globalSettings.tabInsertsSpaces:
            cursor.insertText(' ' * indent)
        else:
            cursor.insertText('\t')

def documentIndentLess(document, cursor, globalSettings=globalSettings):
    if cursor.hasSelection():
        block = document.findBlock(cursor.selectionStart())
        end = document.findBlock(cursor.selectionEnd()).next()
    else:
        block = document.findBlock(cursor.position())
        end = block.next()
    cursor.beginEditBlock()
    while block != end:
        cursor.setPosition(block.position())
        if document.characterAt(cursor.position()) == '\t':
            cursor.deleteChar()
        else:
            pos = 0
            while document.characterAt(cursor.position()) == ' ' \
            and pos < globalSettings.tabWidth:
                pos += 1
                cursor.deleteChar()
        block = block.next()
    cursor.endEditBlock()

class ReTextEdit(QTextEdit):
    resized = pyqtSignal(QRect)
    scrollLimitReached = pyqtSignal(QWheelEvent)
    returnBlockPattern = re.compile("^[\\s]*([*>-]|\\d+\\.) ")
    orderedListPattern = re.compile("^([\\s]*)(\\d+)\\. $")
    wordPattern = re.compile(r"\w+")
    nonAlphaNumPattern = re.compile(r"\W")
    surroundKeysSelfClose = [
        Qt.Key.Key_Underscore,
        Qt.Key.Key_Asterisk,
        Qt.Key.Key_QuoteDbl,
        Qt.Key.Key_QuoteLeft,
        Qt.Key.Key_Apostrophe
    ]
    surroundKeysOtherClose = {
        Qt.Key.Key_ParenLeft: ')',
        Qt.Key.Key_BracketLeft: ']'
    }

    def __init__(self, parent, settings=globalSettings):
        QTextEdit.__init__(self)
        self.tab = weakref.proxy(parent)
        self.parent = parent.p
        self.undoRedoActive = False
        self.tableModeEnabled = False
        self.setAcceptRichText(False)
        self.lineNumberArea = LineNumberArea(self)
        self.infoArea = LineInfoArea(self)
        self.statistics = (0, 0, 0)
        self.statsArea = TextInfoArea(self)
        self.updateFont()
        self.setWrapModeAndWidth()
        self.document().blockCountChanged.connect(self.updateLineNumberAreaWidth)
        self.cursorPositionChanged.connect(self.highlightCurrentLine)
        self.document().contentsChange.connect(self.contentsChange)
        self.settings = settings
        if globalSettings.useFakeVim:
            self.installFakeVimHandler()
        self.mimeDatabase = QMimeDatabase()

    def setWrapModeAndWidth(self):
        if globalSettings.rightMarginWrap:
            self.setLineWrapMode(QTextEdit.LineWrapMode.FixedColumnWidth)
            self.setLineWrapColumnOrWidth(globalSettings.rightMargin or 80)
        else:
            self.setLineWrapMode(QTextEdit.LineWrapMode.WidgetWidth)
        self.setWordWrapMode(QTextOption.WrapMode.WrapAtWordBoundaryOrAnywhere)

    def updateFont(self):
        self.setFont(globalSettings.getEditorFont())
        metrics = self.fontMetrics()
        self.marginx = (int(self.document().documentMargin())
            + metrics.horizontalAdvance(' ' * globalSettings.rightMargin))
        self.setTabStopDistance(globalSettings.tabWidth * metrics.horizontalAdvance(' '))
        self.updateLineNumberAreaWidth()
        self.infoArea.updateTextAndGeometry()
        self.updateTextStatistics()
        self.statsArea.updateTextAndGeometry()
        if globalSettings.wideCursor:
            self.setCursorWidth(metrics.averageCharWidth())

    def paintEvent(self, event):
        if not globalSettings.rightMargin:
            return super().paintEvent(event)
        painter = QPainter(self.viewport())
        painter.setPen(getColor('marginLine'))
        y1 = self.rect().topLeft().y()
        y2 = self.rect().bottomLeft().y()
        painter.drawLine(self.marginx, y1, self.marginx, y2)
        super().paintEvent(event)

    def wheelEvent(self, event):
        modifiers = QGuiApplication.keyboardModifiers()
        if modifiers == Qt.KeyboardModifier.ControlModifier:
            font = globalSettings.getEditorFont()
            size = font.pointSize()
            scroll = event.angleDelta().y()
            if scroll > 0:
                size += 1
            elif scroll < 0:
                size -= 1
            else:
                return
            font.setPointSize(size)
            self.parent.setEditorFont(font)
        else:
            super().wheelEvent(event)

            if event.angleDelta().y() < 0:
                scrollBarLimit = self.verticalScrollBar().maximum()
            else:
                scrollBarLimit = self.verticalScrollBar().minimum()

            if self.verticalScrollBar().value() == scrollBarLimit:
                self.scrollLimitReached.emit(event)

    def scrollContentsBy(self, dx, dy):
        super().scrollContentsBy(dx, dy)
        self.lineNumberArea.update()

    def contextMenuEvent(self, event):
        # Create base menu
        menu = self.createStandardContextMenu()
        if self.parent.actionPasteImage.isEnabled():
            actions = menu.actions()
            actionPaste = menu.findChild(QAction, "edit-paste")
            actionNextAfterPaste = actions[actions.index(actionPaste) + 1]
            menu.insertAction(actionNextAfterPaste, self.parent.actionPasteImage)

        text = self.toPlainText()
        if not text:
            menu.exec(event.globalPos())
            return

        # Check word under the cursor
        oldcursor = self.textCursor()
        cursor = self.cursorForPosition(event.pos())
        curchar = self.document().characterAt(cursor.position())
        isalpha = curchar.isalpha()
        word = None
        if isalpha and not (
            oldcursor.hasSelection()
            and oldcursor.selectedText() != cursor.selectedText()
        ):
            cursor.select(QTextCursor.SelectionType.WordUnderCursor)
            word = cursor.selectedText()

        correct = True
        if word is not None and self.tab.highlighter.dictionaries:
            correct = any(dictionary.check(word) for dictionary in self.tab.highlighter.dictionaries)

        if not correct:
            self.setTextCursor(cursor)
            suggestions = self.tab.highlighter.dictionaries[0].suggest(word)
            actions = [self.parent.act(sug, trig=self.fixWord(sug)) for sug in suggestions]
            menu.insertSeparator(menu.actions()[0])
            for action in actions[::-1]:
                menu.insertAction(menu.actions()[0], action)
            menu.insertSeparator(menu.actions()[0])
            menu.insertAction(
                menu.actions()[0],
                self.parent.act(self.tr('Add to dictionary'), trig=self.learnWord(word)),
            )

        menu.addSeparator()
        menu.addAction(self.parent.actionMoveUp)
        menu.addAction(self.parent.actionMoveDown)

        menu.exec(event.globalPos())

    def fixWord(self, correctword):
        return lambda: self.insertPlainText(correctword)

    def learnWord(self, newword):
        return lambda: self.addNewWord(newword)

    def addNewWord(self, newword):
        cursor = self.textCursor()
        block = cursor.block()
        cursor.clearSelection()
        self.setTextCursor(cursor)
        if self.tab.highlighter.dictionaries and newword:
            dictionary = self.tab.highlighter.dictionaries[0]
            dictionary.add(newword)
            self.tab.highlighter.rehighlightBlock(block)

    def isSurroundKey(self, key):
        return key in self.surroundKeysSelfClose or key in self.surroundKeysOtherClose

    def getCloseKey(self, event, key):
        if key in self.surroundKeysSelfClose:
            return event.text()

        if key in self.surroundKeysOtherClose:
            return self.surroundKeysOtherClose[key]

    def surroundText(self, cursor, event, key):
        text = cursor.selectedText()
        keyStr = event.text()
        keyClose = self.getCloseKey(event, key)

        cursor.insertText(keyStr + text + keyClose)

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = event.modifiers()
        cursor = self.textCursor()
        if key == Qt.Key.Key_Backspace and modifiers & Qt.KeyboardModifier.GroupSwitchModifier:
            # Workaround for https://bugreports.qt.io/browse/QTBUG-49771
            event = QKeyEvent(event.type(), event.key(),
                modifiers ^ Qt.KeyboardModifier.GroupSwitchModifier)
        if key == Qt.Key.Key_Tab:
            documentIndentMore(self.document(), cursor)
        elif key == Qt.Key.Key_Backtab:
            documentIndentLess(self.document(), cursor)
        elif key == Qt.Key.Key_Return:
            markupClass = self.tab.getActiveMarkupClass()
            if modifiers & Qt.KeyboardModifier.ControlModifier:
                cursor.insertText('\n')
                self.ensureCursorVisible()
            elif self.tableModeEnabled and tablemode.handleReturn(cursor, markupClass,
                    newRow=(modifiers & Qt.KeyboardModifier.ShiftModifier)):
                self.setTextCursor(cursor)
                self.ensureCursorVisible()
            else:
                if modifiers & Qt.KeyboardModifier.ShiftModifier and markupClass == MarkdownMarkup:
                    # Insert Markdown-style line break
                    cursor.insertText('  ')
                self.handleReturn(cursor)
        elif key == Qt.Key.Key_Up:
            mode = QTextCursor.MoveMode.MoveAnchor
            if modifiers & Qt.KeyboardModifier.ShiftModifier:
                mode = QTextCursor.MoveMode.KeepAnchor
            oldPos = cursor.position()
            self.moveCursor(QTextCursor.MoveOperation.Up, mode)
            if self.textCursor().position() == oldPos:
                self.moveCursor(QTextCursor.MoveOperation.Start, mode)
        elif key == Qt.Key.Key_Down:
            mode = QTextCursor.MoveMode.MoveAnchor
            if modifiers & Qt.KeyboardModifier.ShiftModifier:
                mode = QTextCursor.MoveMode.KeepAnchor
            oldPos = cursor.position()
            self.moveCursor(QTextCursor.MoveOperation.Down, mode)
            if self.textCursor().position() == oldPos:
                self.moveCursor(QTextCursor.MoveOperation.End, mode)
        elif cursor.selectedText() and self.isSurroundKey(key):
            self.surroundText(cursor, event, key)
        else:
            if event.text() and self.tableModeEnabled:
                cursor.beginEditBlock()
            super().keyPressEvent(event)
            if event.text() and self.tableModeEnabled:
                cursor.endEditBlock()

    def handleReturn(self, cursor):
        # Select text between the cursor and the line start
        cursor.movePosition(QTextCursor.MoveOperation.StartOfBlock, QTextCursor.MoveMode.KeepAnchor)
        text = cursor.selectedText()
        length = len(text)
        match = self.returnBlockPattern.search(text)
        if match is not None:
            matchedText = match.group(0)
            if len(matchedText) == length:
                cursor.removeSelectedText()
                matchedText = ''
            else:
                matchOL = self.orderedListPattern.match(matchedText)
                if matchOL is not None:
                    matchedPrefix = matchOL.group(1)
                    matchedNumber = int(matchOL.group(2))
                    nextNumber = (
                        matchedNumber
                        if self.settings.orderedListMode == 'repeat'
                        else matchedNumber + 1
                    )
                    matchedText = matchedPrefix + str(nextNumber) + ". "
        else:
            matchedText = ''
        # Reset the cursor
        cursor = self.textCursor()
        cursor.insertText('\n' + matchedText)
        self.ensureCursorVisible()

    def moveLineUp(self):
        self.moveLine(QTextCursor.MoveOperation.PreviousBlock)

    def moveLineDown(self):
        self.moveLine(QTextCursor.MoveOperation.NextBlock)

    def moveLine(self, direction):
        cursor = self.textCursor()
        # Select the current block
        cursor.movePosition(QTextCursor.MoveOperation.StartOfBlock, QTextCursor.MoveMode.MoveAnchor)
        cursor.movePosition(QTextCursor.MoveOperation.NextBlock, QTextCursor.MoveMode.KeepAnchor)
        text = cursor.selectedText()
        # Remove it
        cursor.removeSelectedText()
        # Move to the wanted block
        cursor.movePosition(direction, QTextCursor.MoveMode.MoveAnchor)
        # Paste the line
        cursor.insertText(text)
        # Move to the pasted block
        cursor.movePosition(QTextCursor.MoveOperation.PreviousBlock, QTextCursor.MoveMode.MoveAnchor)
        # Update cursor
        self.setTextCursor(cursor)

    def lineNumberAreaWidth(self):
        if not globalSettings.lineNumbersEnabled:
            return 0
        cursor = QTextCursor(self.document())
        cursor.movePosition(QTextCursor.MoveOperation.End)
        if globalSettings.relativeLineNumbers:
            digits = len(str(cursor.blockNumber())) + 1
        else:
            digits = len(str(cursor.blockNumber() + 1))
        return 5 + self.fontMetrics().horizontalAdvance('9') * digits

    def updateLineNumberAreaWidth(self, blockcount=0):
        self.setViewportMargins(self.lineNumberAreaWidth(), 0, 0, 0)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        rect = self.contentsRect()
        self.resized.emit(rect)
        self.lineNumberArea.setGeometry(rect.left(), rect.top(),
            self.lineNumberAreaWidth(), rect.height())
        self.infoArea.updateTextAndGeometry()
        self.statsArea.updateTextAndGeometry()
        self.setWrapModeAndWidth()
        self.ensureCursorVisible()

    def highlightCurrentLine(self):
        if globalSettings.relativeLineNumbers:
            self.lineNumberArea.update()
        if globalSettings.highlightCurrentLine == 'disabled':
            return self.setExtraSelections([])
        selection = QTextEdit.ExtraSelection()
        selection.format.setBackground(getColor('currentLineHighlight'))
        selection.format.setProperty(QTextFormat.Property.FullWidthSelection, True)
        selection.cursor = self.textCursor()
        selection.cursor.clearSelection()
        selections = [selection]
        if globalSettings.highlightCurrentLine == 'wrapped-line':
            selections.append(QTextEdit.ExtraSelection())
            selections[0].cursor.movePosition(QTextCursor.MoveOperation.StartOfBlock)
            selections[0].cursor.movePosition(
                QTextCursor.MoveOperation.EndOfBlock,
                QTextCursor.MoveMode.KeepAnchor,
            )
            selections[1].format.setBackground(getColor('currentLineHighlight'))
            selections[1].format.setProperty(QTextFormat.Property.FullWidthSelection, True)
            selections[1].cursor = self.textCursor()
            selections[1].cursor.movePosition(QTextCursor.MoveOperation.EndOfBlock)
        elif selection.cursor.block().textDirection() == Qt.LayoutDirection.RightToLeft:
            # FullWidthSelection does not work correctly for RTL direction
            selection.cursor.movePosition(QTextCursor.MoveOperation.StartOfLine)
            selection.cursor.movePosition(
                QTextCursor.MoveOperation.EndOfLine,
                QTextCursor.MoveMode.KeepAnchor,
            )
        self.setExtraSelections(selections)

    def enableTableMode(self, enable):
        self.tableModeEnabled = enable

    def backupCursorPositionOnLine(self):
        return self.textCursor().positionInBlock()

    def restoreCursorPositionOnLine(self, positionOnLine):
        cursor = self.textCursor()
        cursor.setPosition(cursor.block().position() + positionOnLine)
        self.setTextCursor(cursor)

    def contentsChange(self, pos, removed, added):
        if self.tableModeEnabled:
            markupClass = self.tab.getActiveMarkupClass()

            cursorPosition = self.backupCursorPositionOnLine()
            tablemode.adjustTableToChanges(self.document(), pos, added - removed, markupClass)
            self.restoreCursorPositionOnLine(cursorPosition)
        self.lineNumberArea.update()
        self.updateTextStatistics()

    def findNextImageName(self, filenames):
        highestNumber = 0
        for filename in filenames:
            m = re.match(r'image(\d+).png', filename, re.IGNORECASE)
            if m:
                number = int(m.group(1))
                highestNumber = max(number, highestNumber)
        return f'image{highestNumber + 1:04d}.png'

    def getImageFilename(self):
        if self.tab.fileName:
            saveDir = os.path.dirname(self.tab.fileName)
        else:
            saveDir = os.getcwd()

        imageFileName = self.findNextImageName(os.listdir(saveDir))

        return QFileDialog.getSaveFileName(self,
                                           self.tr('Save image'),
                                           os.path.join(saveDir, imageFileName),
                                           self.tr('Images (*.png *.jpg)'))[0]

    def makeFileNameRelative(self, fileName):
        """Tries to make the given fileName relative. If the document is
        not saved, or the fileName is on a different root, returns the
        original fileName.
        """
        if self.tab.fileName:
            currentDir = os.path.dirname(self.tab.fileName)
            try:
                return os.path.relpath(fileName, currentDir)
            except ValueError:  # different roots
                return fileName
        return fileName

    def getImageMarkup(self, fileName):
        """Returns markup for image in the current markup language.

        This method is also accessed in ReTextWindow.insertImage.
        """
        link = self.makeFileNameRelative(fileName)
        markupClass = self.tab.getActiveMarkupClass()
        if markupClass == MarkdownMarkup:
            return f'![{QFileInfo(link).baseName()}]({link})'
        elif markupClass == ReStructuredTextMarkup:
            return f'.. image:: {link}'
        elif markupClass == TextileMarkup:
            return f'!{link}!'
        elif markupClass == AsciiDocMarkup:
            return f'image::{link}[]'

    def pasteImage(self):
        mimeData = QApplication.instance().clipboard().mimeData()
        fileName = self.getImageFilename()
        if not fileName or not mimeData.hasImage():
            return
        image = QImage(mimeData.imageData())
        image.save(fileName)

        imageText = self.getImageMarkup(fileName)

        self.textCursor().insertText(imageText)

    def installFakeVimHandler(self):
        if ReTextFakeVimHandler:
            fakeVimEditor = ReTextFakeVimHandler(self, self.parent)
            fakeVimEditor.setSaveAction(self.parent.actionSave)
            fakeVimEditor.setQuitAction(self.parent.actionQuit)
            self.parent.actionFakeVimMode.triggered.connect(fakeVimEditor.remove)

    def updateTextStatistics(self):
        if not globalSettings.documentStatsEnabled:
            return
        text = self.toPlainText()
        wordCount = len(self.wordPattern.findall(text))
        alphaNums = self.nonAlphaNumPattern.sub('', text)
        alphaNumCount = len(alphaNums)
        characterCount = len(text)
        self.statistics = (wordCount, alphaNumCount, characterCount)

    def inputMethodEvent(self, event):
        # Empty events cause layout issues (issue #597), so ignore them.
        if event.preeditString() or event.commitString() or event.attributes():
            super().inputMethodEvent(event)

    def insertFromMimeData(self, source):
        if source.hasUrls():
            url = source.urls()[0]
            mimeType = self.mimeDatabase.mimeTypeForUrl(url)
            if mimeType.name().startswith("image/"):
                if url.isLocalFile():
                    url = url.toLocalFile()
                imageText = self.getImageMarkup(url)
                self.textCursor().insertText(imageText)
                return
        return super().insertFromMimeData(source)


class LineNumberArea(QWidget):
    def __init__(self, editor):
        QWidget.__init__(self, editor)
        self.editor = editor

    def sizeHint(self):
        return QSize(self.editor.lineNumberAreaWidth(), 0)

    def paintEvent(self, event):
        if not globalSettings.lineNumbersEnabled:
            return super().paintEvent(event)
        painter = QPainter(self)
        painter.fillRect(event.rect(), getColor('lineNumberArea'))
        painter.setPen(getColor('lineNumberAreaText'))
        cursor = self.editor.cursorForPosition(QPoint(0, 0))
        atEnd = False
        fontHeight = self.fontMetrics().height()
        height = self.editor.height()
        if globalSettings.relativeLineNumbers:
            relativeTo = self.editor.textCursor().blockNumber()
        else:
            relativeTo = -1
        while not atEnd:
            rect = self.editor.cursorRect(cursor)
            if rect.top() >= height:
                break
            number = str(cursor.blockNumber() - relativeTo).replace('-', '−')
            painter.drawText(0, rect.top(), self.width() - 2,
                             fontHeight, Qt.AlignmentFlag.AlignRight, number)
            cursor.movePosition(QTextCursor.MoveOperation.EndOfBlock)
            atEnd = cursor.atEnd()
            if not atEnd:
                cursor.movePosition(QTextCursor.MoveOperation.NextBlock)

class InfoArea(QLabel):
    def __init__(self, editor, baseColor):
        QWidget.__init__(self, editor)
        self.editor = editor
        self.editor.cursorPositionChanged.connect(self.updateTextAndGeometry)
        self.updateTextAndGeometry()
        self.setAutoFillBackground(True)
        self.baseColor = baseColor
        palette = self.palette()
        palette.setColor(QPalette.ColorRole.Window, self.baseColor)
        self.setPalette(palette)
        self.setCursor(Qt.CursorShape.IBeamCursor)

    def updateTextAndGeometry(self):
        text = self.getText()
        (w, h) = self.getAreaSize(text)
        (x, y) = self.getAreaPosition(w, h)
        self.setText(text)
        self.resize(w, h)
        self.move(x, y)
        self.setVisible(not globalSettings.useFakeVim)

    def getAreaSize(self, text):
        metrics = self.fontMetrics()
        width = metrics.horizontalAdvance(text)
        height = metrics.height()
        return width, height

    def getAreaPosition(self, width, height):
        return 0, 0

    def getText(self):
        return ""

    def enterEvent(self, event):
        palette = self.palette()
        windowColor = QColor(self.baseColor)
        windowColor.setAlpha(0x20)
        palette.setColor(QPalette.ColorRole.Window, windowColor)
        textColor = palette.color(QPalette.ColorRole.WindowText)
        textColor.setAlpha(0x20)
        palette.setColor(QPalette.ColorRole.WindowText, textColor)
        self.setPalette(palette)

    def leaveEvent(self, event):
        palette = self.palette()
        palette.setColor(QPalette.ColorRole.Window, self.baseColor)
        palette.setColor(QPalette.ColorRole.WindowText,
            self.editor.palette().color(QPalette.ColorRole.WindowText))
        self.setPalette(palette)

    def mousePressEvent(self, event):
        pos: QPointF = self.mapToParent(event.position())
        pos.setX(pos.x() - self.editor.lineNumberAreaWidth())
        newEvent = QMouseEvent(event.type(), pos,
                               event.button(), event.buttons(),
                               event.modifiers())
        self.editor.mousePressEvent(newEvent)

    mouseReleaseEvent = mousePressEvent
    mouseDoubleClickEvent = mousePressEvent
    mouseMoveEvent = mousePressEvent

class LineInfoArea(InfoArea):
    def __init__(self, editor):
        InfoArea.__init__(self, editor, getColor('infoArea'))

    def getAreaPosition(self, width, height):
        viewport = self.editor.viewport()
        rightSide = viewport.width() + self.editor.lineNumberAreaWidth()
        if globalSettings.documentStatsEnabled:
            return rightSide - width, viewport.height() - (2 * height)
        else:
            return rightSide - width, viewport.height() - height

    def getText(self):
        template = '%d : %d'
        cursor = self.editor.textCursor()
        block = cursor.blockNumber() + 1
        position = cursor.positionInBlock()
        return template % (block, position)


class TextInfoArea(InfoArea):
    def __init__(self, editor):
        InfoArea.__init__(self, editor, getColor('statsArea'))

    def getAreaPosition(self, width, height):
        viewport = self.editor.viewport()
        rightSide = viewport.width() + self.editor.lineNumberAreaWidth()
        return rightSide - width, viewport.height() - height

    def getText(self):
        if not globalSettings.documentStatsEnabled:
            return
        template = self.tr('%d w | %d a | %d c',
                           'count of words, alphanumeric characters, all characters')
        words, alphaNums, characters = self.editor.statistics
        return template % (words, alphaNums, characters)



# vim: ts=4:sw=4:expandtab
#
# This file is part of ReText
# Copyright: 2015-2025 Dmitry Shachnev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import locale
import time
from os.path import exists, splitext

from markups import find_markup_class_by_name, get_markup_for_file_name
from markups.common import MODULE_HOME_PAGE
from PyQt6.QtCore import QDir, QFile, QFileInfo, QPoint, Qt, QTimer, QUrl, pyqtSignal
from PyQt6.QtGui import QTextCursor, QTextDocument
from PyQt6.QtWidgets import QMessageBox, QSplitter, QTextEdit

from ReText import app_version, converterprocess, globalSettings
from ReText.editor import ReTextEdit
from ReText.highlighter import ReTextHighlighter
from ReText.preview import ReTextPreview

try:
    from ReText.webenginepreview import ReTextWebEnginePreview
except ImportError:
    ReTextWebEnginePreview = None

PreviewDisabled, PreviewLive, PreviewNormal = range(3)

class ReTextTab(QSplitter):

    fileNameChanged = pyqtSignal()
    modificationStateChanged = pyqtSignal()
    activeMarkupChanged = pyqtSignal()

    # Make _fileName a read-only property to make sure that any
    # modification happens through the proper functions. These functions
    # will make sure that the fileNameChanged signal is emitted when
    # applicable.
    @property
    def fileName(self):
        return self._fileName

    def __init__(self, parent, fileName, previewState=PreviewDisabled):
        super().__init__(Qt.Orientation.Horizontal, parent=parent)
        self.p = parent
        self._fileName = fileName
        self.editBox = ReTextEdit(self)
        self.previewBox = self.createPreviewBox(self.editBox)
        self.activeMarkupClass = None
        self.markup = None
        self.converted = None
        self.previewState = previewState
        self.previewOutdated = False
        self.conversionPending = False
        self.cssFileExists = False
        self.forceDisableAutoSave = False

        self.converterProcess = converterprocess.ConverterProcess()
        self.converterProcess.conversionDone.connect(self.updatePreviewBox)

        textDocument = self.editBox.document()
        self.highlighter = ReTextHighlighter(textDocument)
        if parent.actionEnableSC.isChecked():
            dictionaries, _errors = parent.getSpellCheckDictionaries()
            self.highlighter.dictionaries = dictionaries
            # Rehighlighting is tied to the change in markup class that
            # happens at the end of this function

        self.editBox.textChanged.connect(self.triggerPreviewUpdate)
        self.editBox.undoAvailable.connect(parent.actionUndo.setEnabled)
        self.editBox.redoAvailable.connect(parent.actionRedo.setEnabled)
        self.editBox.copyAvailable.connect(parent.enableCopy)

        # Give both boxes a minimum size so the minimumSizeHint will be
        # ignored when splitter.setSizes is called below
        for widget in self.editBox, self.previewBox:
            widget.setMinimumWidth(125)
            self.addWidget(widget)
        self.setSizes((50, 50))
        self.setChildrenCollapsible(False)

        textDocument.modificationChanged.connect(self.handleModificationChanged)

        self.updateActiveMarkupClass()

    def handleModificationChanged(self):
        self.modificationStateChanged.emit()

    def createPreviewBox(self, editBox):

        # Use closures to avoid a hard reference from ReTextWebEnginePreview
        # to self, which would keep the tab and its resources alive
        # even after other references to it have disappeared.

        def editorPositionToSourceLine(editorPosition):
            viewportPosition = editorPosition - editBox.verticalScrollBar().value()
            sourceLine = editBox.cursorForPosition(QPoint(0,viewportPosition)).blockNumber()
            return sourceLine

        def sourceLineToEditorPosition(sourceLine):
            doc = editBox.document()
            block = doc.findBlockByNumber(sourceLine)
            rect = doc.documentLayout().blockBoundingRect(block)
            return rect.top()

        if ReTextWebEnginePreview and globalSettings.useWebEngine:
            preview = ReTextWebEnginePreview(self,
                                             editorPositionToSourceLine,
                                             sourceLineToEditorPosition)
        else:
            preview = ReTextPreview(self)

        return preview

    def getActiveMarkupClass(self):
        '''
        Return the currently active markup class for this tab.
        No objects should be created of this class, it should
        only be used to retrieve markup class specific information.
        '''
        return self.activeMarkupClass

    def updateActiveMarkupClass(self):
        '''
        Update the active markup class based on the default class and
        the current filename. If the active markup class changes, the
        highlighter is rerun on the input text, the markup object of
        this tab is replaced with one of the new class and the
        activeMarkupChanged signal is emitted.
        '''
        previousMarkupClass = self.activeMarkupClass

        self.activeMarkupClass = find_markup_class_by_name(globalSettings.defaultMarkup)

        if self._fileName:
            markupClass = get_markup_for_file_name(
                self._fileName, return_class=True)
            if markupClass:
                self.activeMarkupClass = markupClass

        if self.activeMarkupClass != previousMarkupClass:
            self.highlighter.docType = self.activeMarkupClass.name if self.activeMarkupClass else None
            self.highlighter.rehighlight()

            self.activeMarkupChanged.emit()
            self.triggerPreviewUpdate()

    def getDocumentTitleFromConverted(self, converted):
        if converted:
            try:
                return converted.get_document_title()
            except Exception:
                self.p.printError()

        return self.getBaseName()

    def getBaseName(self):
        if self._fileName:
            fileinfo = QFileInfo(self._fileName)
            basename = fileinfo.completeBaseName()
            return (basename if basename else fileinfo.fileName())
        return self.tr("New document")

    def getHtmlFromConverted(self, converted, includeStyleSheet=True, webenv=False):
        if converted is None:
            markupClass = self.getActiveMarkupClass()
            errMsg = self.tr('Could not parse file contents, check if '
                             'you have the <a href="%s">necessary module</a> '
                             'installed!')
            try:
                errMsg %= markupClass.attributes[MODULE_HOME_PAGE]
            except (AttributeError, KeyError):
                # Remove the link if markupClass doesn't have the needed attribute
                errMsg = errMsg.replace('<a href="%s">', '').replace('</a>', '')
            return f'<p style="color: red">{errMsg}</p>'
        headers = ''
        if includeStyleSheet and self.p.ss is not None:
            headers += '<style type="text/css">\n' + self.p.ss + '</style>\n'
        elif includeStyleSheet:
            style = 'td, th { border: 1px solid #c3c3c3; padding: 0 3px 0 3px; }\n'
            style += 'table { border-collapse: collapse; }\n'
            style += 'img { max-width: 100%; }\n'
            headers += '<style type="text/css">\n' + style + '</style>\n'
        baseName = self.getBaseName()
        if self.cssFileExists:
            headers += f'<link rel="stylesheet" type="text/css" href="{baseName}.css">\n'
        headers += f'<meta name="generator" content="ReText {app_version}">\n'
        return converted.get_whole_html(
            custom_headers=headers, include_stylesheet=includeStyleSheet,
            fallback_title=baseName, webenv=webenv)

    def getDocumentForExport(self, includeStyleSheet=True, webenv=False):
        markupClass = self.getActiveMarkupClass()
        if markupClass and markupClass.available():
            exportMarkup = markupClass(filename=self._fileName)

            text = self.editBox.toPlainText()
            converted = exportMarkup.convert(text)
        else:
            converted = None

        return (self.getDocumentTitleFromConverted(converted),
                self.getHtmlFromConverted(converted, includeStyleSheet=includeStyleSheet, webenv=webenv),
            self.previewBox)

    def updatePreviewBox(self):
        self.conversionPending = False

        try:
            self.converted = self.converterProcess.get_result()
        except converterprocess.MarkupNotAvailableError:
            self.converted = None
        except converterprocess.ConversionError:
            return self.p.printError()

        if isinstance(self.previewBox, QTextEdit):
            scrollbar = self.previewBox.verticalScrollBar()
            scrollbarValue = scrollbar.value()
            # If scrollbar was not on top, save its distance to bottom so that
            # it will be restored in previewBox.updateScrollPosition() later.
            if scrollbarValue:
                self.previewBox.distToBottom = scrollbar.maximum() - scrollbarValue
            else:
                self.previewBox.distToBottom = None
        try:
            html = self.getHtmlFromConverted(self.converted)
        except Exception:
            return self.p.printError()
        self.previewBox.setFont(globalSettings.getPreviewFont())
        if isinstance(self.previewBox, QTextEdit):
            self.previewBox.lastRenderTime = time.time()
            self.previewBox.setHtml(html)
            self.previewBox.updateScrollPosition(scrollbar.minimum(),
                                                 scrollbar.maximum())
        else:
            # Always provide a baseUrl otherwise QWebView will
            # refuse to show images or other external objects
            if self._fileName:
                baseUrl = QUrl.fromLocalFile(self._fileName)
            else:
                baseUrl = QUrl.fromLocalFile(QDir.currentPath())
            self.previewBox.setHtml(html, baseUrl)

        if self.previewOutdated:
            self.triggerPreviewUpdate()

    def triggerPreviewUpdate(self):
        self.previewOutdated = True
        if self.previewState == PreviewDisabled:
            return

        if not self.conversionPending:
            self.conversionPending = True
            QTimer.singleShot(500, self.startPendingConversion)

    def startPendingConversion(self):
        self.previewOutdated = False

        requested_extensions = ['ReText.mdx_posmap'] if globalSettings.syncScroll else []
        self.converterProcess.start_conversion(self.getActiveMarkupClass().name,
                                               self.fileName,
                                               requested_extensions,
                                               self.editBox.toPlainText(),
                                               QDir.currentPath())

    def updateBoxesVisibility(self):
        self.editBox.setVisible(self.previewState < PreviewNormal)
        self.previewBox.setVisible(self.previewState > PreviewDisabled)

    def rebuildPreviewBox(self):
        self.previewBox.disconnectExternalSignals()
        self.previewBox.setParent(None)
        self.previewBox.deleteLater()
        self.previewBox = self.createPreviewBox(self.editBox)
        self.previewBox.setMinimumWidth(125)
        self.addWidget(self.previewBox)
        self.setSizes((50, 50))
        self.triggerPreviewUpdate()
        self.updateBoxesVisibility()

    def detectEncoding(self, raw: bytes):
        '''
        Detect content encoding of the given data.

        It will return None if it can't determine the encoding.
        '''
        try:
            import chardet
        except ImportError:
            return

        result = chardet.detect(raw)
        if result['confidence'] > 0.9:
            if result['encoding'].lower() == 'ascii':
                # UTF-8 files can be falsely detected as ASCII files if they
                # don't contain non-ASCII characters in first 2048 bytes.
                # We map ASCII to UTF-8 to avoid such situations.
                return 'utf-8'
            return result['encoding']

    def readTextFromFile(self, fileName=None, encoding=None):
        previousFileName = self._fileName
        fileName = fileName or self._fileName

        try:
            with open(fileName, 'rb') as openfile:
                data = openfile.read()
        except OSError as ex:
            QMessageBox.warning(self, '', str(ex))
            return

        # Only try to detect encoding if it is not specified
        if encoding is None and globalSettings.detectEncoding:
            encoding = self.detectEncoding(data[:2048])
        encoding = encoding or globalSettings.defaultCodec or None

        try:
            text = data.decode(encoding or locale.getpreferredencoding(False))
        except (UnicodeDecodeError, LookupError) as ex:
            QMessageBox.warning(self, '', str(ex))
            return

        if encoding:
            # If encoding is specified or detected, we should save the file with
            # the same encoding
            self.editBox.document().setProperty("encoding", encoding)

        self._fileName = fileName
        if previousFileName != self._fileName:
            self.updateActiveMarkupClass()

        self.forceDisableAutoSave = False
        self.editBox.setPlainText(text)
        self.editBox.document().setModified(False)
        self.handleModificationChanged()

        cssFileName = self.getBaseName() + '.css'
        self.cssFileExists = QFile.exists(cssFileName)

        if previousFileName != self._fileName:
            self.fileNameChanged.emit()

    def writeTextToFile(self, fileName=None):
        # Just writes the text to file, without any changes to tab object
        # Used directly for e.g. export extensions

        # Get text from the cursor to avoid tweaking special characters,
        # see https://bugreports.qt.io/browse/QTBUG-57552 and
        # https://github.com/retext-project/retext/issues/216
        cursor = self.editBox.textCursor()
        cursor.select(QTextCursor.SelectionType.Document)
        text = cursor.selectedText().replace('\u2029', '\n')

        fileName = fileName or self._fileName
        encoding = self.editBox.document().property("encoding")
        encoding = encoding or globalSettings.defaultCodec or None
        try:
            data = text.encode(encoding or locale.getpreferredencoding(False))
        except (UnicodeEncodeError, LookupError) as ex:
            QMessageBox.warning(self, '', str(ex))
            return False
        try:
            with open(fileName, 'wb') as savefile:
                savefile.write(data)
        except OSError as ex:
            QMessageBox.warning(self, '', str(ex))
            return False
        return True

    def saveTextToFile(self, fileName=None):
        # Sets fileName as tab fileName and writes the text to that file
        if self._fileName:
            self.p.fileSystemWatcher.removePath(self._fileName)
        result = self.writeTextToFile(fileName)
        if result:
            self.forceDisableAutoSave = False
            self.editBox.document().setModified(False)
            self.p.fileSystemWatcher.addPath(fileName or self._fileName)
            if fileName and self._fileName != fileName:
                self._fileName = fileName
                self.updateActiveMarkupClass()
                self.fileNameChanged.emit()

        return result

    def goToLine(self,line):
        block = self.editBox.document().findBlockByLineNumber(line)
        if block.isValid():
            newCursor = QTextCursor(block)
            self.editBox.setTextCursor(newCursor)

    def find(self, text, flags, replaceText=None, wrap=False):
        if self.previewState == PreviewNormal and replaceText is None:
            return self.previewBox.findText(text, flags)
        cursor = self.editBox.textCursor()
        if wrap and flags & QTextDocument.FindFlag.FindBackward:
            cursor.movePosition(QTextCursor.MoveOperation.End)
        elif wrap:
            cursor.movePosition(QTextCursor.MoveOperation.Start)
        if replaceText is not None and cursor.selectedText() == text:
            newCursor = cursor
        else:
            newCursor = self.editBox.document().find(text, cursor, flags)
        if not newCursor.isNull():
            if replaceText is not None:
                newCursor.insertText(replaceText)
                newCursor.movePosition(
                    QTextCursor.MoveOperation.Left,
                    QTextCursor.MoveMode.MoveAnchor,
                    len(replaceText),
                )
                newCursor.movePosition(
                    QTextCursor.MoveOperation.Right,
                    QTextCursor.MoveMode.KeepAnchor,
                    len(replaceText),
                )
            self.editBox.setTextCursor(newCursor)
            if self.editBox.cursorRect().bottom() >= self.editBox.height() - 3:
                scrollValue = self.editBox.verticalScrollBar().value()
                areaHeight = self.editBox.fontMetrics().height()
                self.editBox.verticalScrollBar().setValue(scrollValue + areaHeight)
            return True
        if not wrap:
            return self.find(text, flags, replaceText, True)
        return False

    def replaceAll(self, text, replaceText):
        cursor = self.editBox.textCursor()
        cursor.beginEditBlock()
        cursor.movePosition(QTextCursor.MoveOperation.Start)
        flags = QTextDocument.FindFlag(0)
        cursor = lastCursor = self.editBox.document().find(text, cursor, flags)
        while not cursor.isNull():
            cursor.insertText(replaceText)
            lastCursor = cursor
            cursor = self.editBox.document().find(text, cursor, flags)
        if not lastCursor.isNull():
            lastCursor.movePosition(
                QTextCursor.MoveOperation.Left,
                QTextCursor.MoveMode.MoveAnchor,
                len(replaceText),
            )
            lastCursor.movePosition(
                QTextCursor.MoveOperation.Right,
                QTextCursor.MoveMode.KeepAnchor,
                len(replaceText),
            )
            self.editBox.setTextCursor(lastCursor)
        self.editBox.textCursor().endEditBlock()
        return not lastCursor.isNull()

    def openSourceFile(self, linkPath):
        """Finds and opens the source file for link target fileToOpen.

        When links like [test](test) are clicked, the file test.md is opened.
        It has to be located next to the current opened file.
        Relative paths like [test](../test) or [test](folder/test) are also possible.
        """

        fileToOpen = self.resolveSourceFile(linkPath)
        if exists(fileToOpen) and get_markup_for_file_name(fileToOpen, return_class=True):
            self.p.openFileWrapper(fileToOpen)
            return fileToOpen
        if get_markup_for_file_name(fileToOpen, return_class=True):
            if not QFile.exists(fileToOpen) and QFileInfo(fileToOpen).dir().exists():
                if self.promptFileCreation(fileToOpen):
                    self.p.openFileWrapper(fileToOpen)
                    return fileToOpen

    def promptFileCreation(self, fileToCreate):
        """
        Prompt user if a file should be created for the clicked link,
        and try to create it. Return True on success.
        """
        buttonReply = QMessageBox.question(
            self,
            self.tr('Create missing file?'),
            self.tr("The file '%s' does not exist.\n\nDo you want to create it?") % fileToCreate,
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            QMessageBox.StandardButton.No,
        )
        if buttonReply == QMessageBox.StandardButton.Yes:
            return self.createFile(fileToCreate)
        elif buttonReply == QMessageBox.StandardButton.No:
            return False

    def resolveSourceFile(self, linkPath):
        """
        Finds the actual path of the file to open in a new tab.
        When the link has no extension, eg: [Test](test), the extension of the current file is assumed
        (eg test.md for a markdown file).
        When the link is an html file eg: [Test](test.html), the extension of the current file is assumed
        (eg test.md for a markdown file).
        Relative paths like [test](../test) or [test](folder/test) are also possible.
        """
        basename, ext = splitext(linkPath)
        if self.fileName:
            currentExt = splitext(self.fileName)[1]
            if ext in ('.html', '') and (exists(basename+currentExt) or not exists(linkPath)):
                ext = currentExt

        return basename+ext

    def createFile(self, fileToCreate):
        """Try to create file, return True if successful"""
        try:
            # Create file:
            open(fileToCreate, 'x').close()
            return True
        except OSError as err:
            QMessageBox.warning(self, self.tr("File could not be created"),
                                self.tr("Could not create file '%s': %s") % (fileToCreate, err))
            return False

    def autoSaveActive(self) -> bool:
        return (globalSettings.autoSave
                and not self.forceDisableAutoSave
                and self.fileName is not None
                and QFileInfo(self.fileName).isWritable())



# vim: ts=4:sw=4:expandtab
#
# This file is part of ReText
# Copyright: 2012-2025 Dmitry Shachnev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import configparser
import os
import shlex
import sys
import traceback
import warnings
from subprocess import Popen

import markups

from ReText import app_version, getBundledIcon, globalCache, globalSettings
from ReText.config import ConfigDialog, setIconThemeFromSettings
from ReText.dialogs import EncodingDialog, HtmlDialog, LocaleDialog
from ReText.filesystemmodel import ReTextFileSystemModel
from ReText.tab import (
    PreviewDisabled,
    PreviewLive,
    PreviewNormal,
    ReTextTab,
    ReTextWebEnginePreview,
)
from ReText.tabledialog import InsertTableDialog

try:
    from ReText.fakevimeditor import FakeVimMode, ReTextFakeVimHandler
except ImportError:
    ReTextFakeVimHandler = None

try:
    import enchant
except ImportError:
    enchant = None

from PyQt6.QtCore import (
    QByteArray,
    QDir,
    QFile,
    QFileInfo,
    QFileSystemWatcher,
    QLocale,
    QMarginsF,
    QStandardPaths,
    Qt,
    QTimer,
    QUrl,
    pyqtSlot,
)
from PyQt6.QtGui import (
    QAction,
    QActionGroup,
    QColor,
    QDesktopServices,
    QIcon,
    QKeySequence,
    QPageLayout,
    QPageSize,
    QPalette,
    QTextDocument,
    QTextDocumentWriter,
)
from PyQt6.QtPrintSupport import QPrintDialog, QPrinter, QPrintPreviewDialog
from PyQt6.QtWidgets import (
    QApplication,
    QCheckBox,
    QComboBox,
    QDialog,
    QFileDialog,
    QFontDialog,
    QInputDialog,
    QLineEdit,
    QMainWindow,
    QMenu,
    QMessageBox,
    QSplitter,
    QTabWidget,
    QToolBar,
    QToolButton,
    QTreeView,
)

previewStatesByName = {
    'editor': PreviewDisabled,
    'normal-preview': PreviewNormal,
    'live-preview': PreviewLive,
}


class ReTextWindow(QMainWindow):
    def __init__(self, parent=None):
        QMainWindow.__init__(self, parent)
        self.resize(950, 700)
        qApp = QApplication.instance()
        screenRect = self.screen().geometry()
        if globalCache.windowGeometry:
            self.restoreGeometry(globalCache.windowGeometry)
        else:
            self.move((screenRect.width() - self.width()) // 2,
                      (screenRect.height() - self.height()) // 2)
            if not screenRect.contains(self.geometry()):
                self.showMaximized()
        if sys.platform.startswith('darwin'):
            # https://github.com/retext-project/retext/issues/198
            searchPaths = QIcon.themeSearchPaths()
            searchPaths.append('/opt/local/share/icons')
            searchPaths.append('/usr/local/share/icons')
            QIcon.setThemeSearchPaths(searchPaths)
        setIconThemeFromSettings()
        if QFile.exists(getBundledIcon('retext')):
            self.setWindowIcon(QIcon(getBundledIcon('retext')))
        elif QFile.exists('/usr/share/pixmaps/retext.png'):
            self.setWindowIcon(QIcon('/usr/share/pixmaps/retext.png'))
        else:
            self.setWindowIcon(QIcon.fromTheme('retext',
                QIcon.fromTheme('accessories-text-editor')))
        self.splitter = QSplitter(self)
        self.treeView = QTreeView(self.splitter)
        self.treeView.doubleClicked.connect(self.treeItemSelected)
        self.initDirectoryTree()
        self.treeView.setVisible(globalSettings.showDirectoryTree)
        self.tabWidget = QTabWidget(self.splitter)
        self.initTabWidget()
        if globalCache.splitterState:
            self.splitter.restoreState(globalCache.splitterState)
        else:
            self.splitter.setSizes([self.width() // 5, self.width() * 4 // 5])
        self.setCentralWidget(self.splitter)
        self.tabWidget.currentChanged.connect(self.changeIndex)
        self.tabWidget.tabCloseRequested.connect(self.closeTab)
        self.toolBar = QToolBar(self.tr('File toolbar'), self)
        self.addToolBar(Qt.ToolBarArea.TopToolBarArea, self.toolBar)
        self.editBar = QToolBar(self.tr('Edit toolbar'), self)
        self.addToolBar(Qt.ToolBarArea.TopToolBarArea, self.editBar)
        self.searchBar = QToolBar(self.tr('Search toolbar'), self)
        self.addToolBar(Qt.ToolBarArea.BottomToolBarArea, self.searchBar)
        self.toolBar.setVisible(not globalSettings.hideToolBar)
        self.editBar.setVisible(not globalSettings.hideToolBar)
        self.actionNew = self.act(self.tr('New'), 'document-new',
            self.createNew, shct=QKeySequence.StandardKey.New)
        self.actionOpen = self.act(self.tr('Open'), 'document-open',
            self.openFile, shct=QKeySequence.StandardKey.Open)
        self.actionSetEncoding = self.act(self.tr('Set encoding'),
            trig=self.showEncodingDialog)
        self.actionSetEncoding.setEnabled(False)
        self.actionReload = self.act(self.tr('Reload'), 'view-refresh',
            lambda: self.currentTab.readTextFromFile())
        self.actionReload.setEnabled(False)
        self.actionSave = self.act(self.tr('Save'), 'document-save',
            self.saveFile, shct=QKeySequence.StandardKey.Save)
        self.actionSave.setEnabled(False)
        self.actionSaveAs = self.act(self.tr('Save as'), 'document-save-as',
            self.saveFileAs, shct=QKeySequence.StandardKey.SaveAs)
        self.actionNextTab = self.act(self.tr('Next tab'), 'go-next',
            lambda: self.switchTab(1))
        self.actionNextTab.setShortcuts([
            *QKeySequence.keyBindings(QKeySequence.StandardKey.NextChild),
            QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_PageDown),
        ])
        self.actionPrevTab = self.act(self.tr('Previous tab'), 'go-previous',
            lambda: self.switchTab(-1))
        self.actionPrevTab.setShortcuts([
            *QKeySequence.keyBindings(QKeySequence.StandardKey.PreviousChild),
            QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_PageUp),
            # https://bugreports.qt.io/browse/QTBUG-15746
            QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Backtab),
        ])
        self.actionCloseCurrentTab = self.act(self.tr('Close tab'), 'window-close',
            lambda: self.closeTab(self.ind), shct=QKeySequence.StandardKey.Close)
        self.actionPrint = self.act(self.tr('Print'), 'document-print',
            self.printFile, shct=QKeySequence.StandardKey.Print)
        self.actionPrintPreview = self.act(self.tr('Print preview'), 'document-print-preview',
            self.printPreview)
        self.actionViewHtml = self.act(
            self.tr('View HTML code'),
            'text-html',
            self.viewHtml,
            shct=Qt.Modifier.CTRL | Qt.Key.Key_H,
        )
        self.actionChangeEditorFont = self.act(self.tr('Change editor font'),
            trig=self.changeEditorFont)
        self.actionChangePreviewFont = self.act(self.tr('Change preview font'),
            trig=self.changePreviewFont)
        self.actionSearch = self.act(self.tr('Find text'), 'edit-find',
            self.search, shct=QKeySequence.StandardKey.Find)
        self.actionGoToLine = self.act(self.tr('Go to line'),
            trig=self.goToLine, shct=Qt.Modifier.CTRL | Qt.Key.Key_G)
        self.searchBar.visibilityChanged.connect(self.searchBarVisibilityChanged)
        self.actionPreview = self.act(self.tr('Preview'), shct=Qt.Modifier.CTRL | Qt.Key.Key_E,
            trigbool=self.preview)
        if QIcon.hasThemeIcon('document-preview'):
            self.actionPreview.setIcon(QIcon.fromTheme('document-preview'))
        elif QIcon.hasThemeIcon('preview-file'):
            self.actionPreview.setIcon(QIcon.fromTheme('preview-file'))
        elif QIcon.hasThemeIcon('x-office-document'):
            self.actionPreview.setIcon(QIcon.fromTheme('x-office-document'))
        else:
            self.actionPreview.setIcon(QIcon(getBundledIcon('document-preview')))
        self.actionLivePreview = self.act(self.tr('Live preview'), shct=Qt.Modifier.CTRL | Qt.Key.Key_L,
        trigbool=self.enableLivePreview)
        menuPreview = QMenu()
        menuPreview.addAction(self.actionLivePreview)
        self.actionInsertTable = self.act(self.tr('Insert table'),
            trig=lambda: self.insertFormatting('table'))
        self.actionTableMode = self.act(self.tr('Table editing mode'),
            shct=Qt.Modifier.CTRL | Qt.Key.Key_T,
            trigbool=lambda x: self.currentTab.editBox.enableTableMode(x))
        self.actionInsertImages = self.act(self.tr('Insert images by file path'),
            trig=lambda: self.insertImages())
        if ReTextFakeVimHandler:
            self.actionFakeVimMode = self.act(self.tr('FakeVim mode'),
                shct=Qt.Modifier.CTRL | Qt.Modifier.ALT | Qt.Key.Key_V, trigbool=self.enableFakeVimMode)
            if globalSettings.useFakeVim:
                self.actionFakeVimMode.setChecked(True)
                self.enableFakeVimMode(True)
        self.actionFullScreen = self.act(self.tr('Fullscreen mode'), 'view-fullscreen',
            shct=QKeySequence.StandardKey.FullScreen, trigbool=self.enableFullScreen)
        self.actionFullScreen.setChecked(self.isFullScreen())
        self.actionConfig = self.act(self.tr('Preferences'), trig=self.openConfigDialog)
        if QIcon.hasThemeIcon('configure'):
            self.actionConfig.setIcon(QIcon.fromTheme('configure'))
        else:
            self.actionConfig.setIcon(QIcon.fromTheme('preferences-system'))
        self.actionConfig.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.actionSaveHtml = self.act('HTML', 'text-html', self.saveFileHtml)
        self.actionPdf = self.act('PDF', 'application-pdf', self.savePdf)
        self.actionOdf = self.act('ODT', 'x-office-document', self.saveOdf)
        self.getExportExtensionsList()
        self.actionQuit = self.act(
            self.tr('Quit'),
            'application-exit',
            shct=QKeySequence.StandardKey.Quit,
        )
        self.actionQuit.setMenuRole(QAction.MenuRole.QuitRole)
        self.actionQuit.triggered.connect(self.close)
        self.actionUndo = self.act(self.tr('Undo'), 'edit-undo',
            lambda: self.currentTab.editBox.undo(), shct=QKeySequence.StandardKey.Undo)
        self.actionRedo = self.act(self.tr('Redo'), 'edit-redo',
            lambda: self.currentTab.editBox.redo(), shct=QKeySequence.StandardKey.Redo)
        self.actionCopy = self.act(self.tr('Copy'), 'edit-copy',
            lambda: self.currentTab.editBox.copy(), shct=QKeySequence.StandardKey.Copy)
        self.actionCut = self.act(self.tr('Cut'), 'edit-cut',
            lambda: self.currentTab.editBox.cut(), shct=QKeySequence.StandardKey.Cut)
        self.actionPaste = self.act(self.tr('Paste'), 'edit-paste',
            lambda: self.currentTab.editBox.paste(), shct=QKeySequence.StandardKey.Paste)
        self.actionPasteImage = self.act(
            self.tr('Paste image'),
            'edit-paste',
            lambda: self.currentTab.editBox.pasteImage(),
            shct=Qt.Modifier.CTRL | Qt.Modifier.SHIFT | Qt.Key.Key_V,
        )
        self.actionMoveUp = self.act(self.tr('Move line up'), 'go-up',
            lambda: self.currentTab.editBox.moveLineUp(), shct=Qt.Modifier.ALT | Qt.Key.Key_Up)
        self.actionMoveDown = self.act(self.tr('Move line down'), 'go-down',
            lambda: self.currentTab.editBox.moveLineDown(), shct=Qt.Modifier.ALT | Qt.Key.Key_Down)
        self.actionUndo.setEnabled(False)
        self.actionRedo.setEnabled(False)
        self.actionCopy.setEnabled(False)
        self.actionCut.setEnabled(False)
        qApp.clipboard().dataChanged.connect(self.clipboardDataChanged)
        self.clipboardDataChanged()
        self.actionEnableSC = self.act(self.tr('Enable'), trigbool=self.enableSpellCheck)
        self.actionSetLocale = self.act(self.tr('Set locale'), trig=self.changeLocale)
        self.actionWebEngine = self.act(self.tr('Use WebEngine (Chromium) renderer'),
            trigbool=self.enableWebEngine)
        if ReTextWebEnginePreview is None:
            globalSettings.useWebEngine = False
            self.actionWebEngine.setEnabled(False)
        self.actionWebEngine.setChecked(globalSettings.useWebEngine)
        self.actionShow = self.act(self.tr('Show directory'), 'document-open-folder', self.showInDir)
        self.actionShowDirectoryTree = self.act(self.tr('Show directory tree'),
            trigbool=self.treeView.setVisible,
            shct=Qt.Key.Key_F9)
        self.actionShowDirectoryTree.setChecked(globalSettings.showDirectoryTree)
        self.actionFind = self.act(self.tr('Next'), 'go-next', self.find,
            shct=QKeySequence.StandardKey.FindNext)
        self.actionFindPrev = self.act(self.tr('Previous'), 'go-previous',
            lambda: self.find(back=True), shct=QKeySequence.StandardKey.FindPrevious)
        self.actionReplace = self.act(self.tr('Replace'), 'edit-find-replace',
            lambda: self.find(replace=True))
        self.actionReplaceAll = self.act(self.tr('Replace all'), trig=self.replaceAll)
        menuReplace = QMenu()
        menuReplace.addAction(self.actionReplaceAll)
        self.actionCloseSearch = self.act(self.tr('Close'), 'window-close',
            lambda: self.searchBar.setVisible(False),
            shct=QKeySequence.StandardKey.Cancel)
        self.actionCloseSearch.setPriority(QAction.Priority.LowPriority)
        self.actionHelp = self.act(self.tr('Get help online'), 'help-contents', self.openHelp)
        self.actionWhatsNew = self.act(self.tr("What's new"), trig=self.openReleases)
        self.aboutWindowTitle = self.tr('About ReText')
        self.actionAbout = self.act(self.aboutWindowTitle, 'help-about', self.aboutDialog)
        self.actionAbout.setMenuRole(QAction.MenuRole.AboutRole)
        self.actionAboutQt = self.act(self.tr('About Qt'))
        self.actionAboutQt.setMenuRole(QAction.MenuRole.AboutQtRole)
        self.actionAboutQt.triggered.connect(qApp.aboutQt)
        availableMarkups = markups.get_available_markups()
        if not availableMarkups:
            print('Warning: no markups are available!')
        if len(availableMarkups) > 1:
            self.chooseGroup = QActionGroup(self)
            markupActions = []
            for markup in availableMarkups:
                markupAction = self.act(markup.name, trigbool=self.markupFunction(markup))
                if markup.name == globalSettings.defaultMarkup:
                    markupAction.setChecked(True)
                self.chooseGroup.addAction(markupAction)
                markupActions.append(markupAction)
        self.actionBold = self.act(self.tr('Bold'), shct=QKeySequence.StandardKey.Bold,
            trig=lambda: self.insertFormatting('bold'))
        self.actionItalic = self.act(self.tr('Italic'), shct=QKeySequence.StandardKey.Italic,
            trig=lambda: self.insertFormatting('italic'))
        self.actionUnderline = self.act(self.tr('Underline'), shct=QKeySequence.StandardKey.Underline,
            trig=lambda: self.insertFormatting('underline'))
        self.usefulTags = ('header', 'italic', 'bold', 'underline', 'numbering',
            'bullets', 'image', 'link', 'inline code', 'code block', 'blockquote',
            'table')
        self.usefulChars = ('deg', 'divide', 'euro', 'hellip', 'laquo', 'larr',
            'lsquo', 'mdash', 'middot', 'minus', 'nbsp', 'ndash', 'raquo',
            'rarr', 'rsquo', 'times')
        self.formattingBox = QComboBox(self.editBar)
        self.formattingBox.addItem(self.tr('Formatting'))
        self.formattingBox.addItems(self.usefulTags)
        self.formattingBox.textActivated.connect(self.insertFormatting)
        self.symbolBox = QComboBox(self.editBar)
        self.symbolBox.addItem(self.tr('Symbols'))
        self.symbolBox.addItems(self.usefulChars)
        self.symbolBox.activated.connect(self.insertSymbol)
        self.updateStyleSheet()
        menubar = self.menuBar()
        menuFile = menubar.addMenu(self.tr('&File'))
        menuEdit = menubar.addMenu(self.tr('&Edit'))
        menuHelp = menubar.addMenu(self.tr('&Help'))
        menuFile.addAction(self.actionNew)
        menuFile.addAction(self.actionOpen)
        self.menuRecentFiles = menuFile.addMenu(
            self.actIcon('document-open-recent'),
            self.tr('Open recent'),
        )
        self.menuRecentFiles.aboutToShow.connect(self.updateRecentFiles)
        menuFile.addAction(self.actionShow)
        menuFile.addAction(self.actionShowDirectoryTree)
        menuFile.addAction(self.actionSetEncoding)
        menuFile.addAction(self.actionReload)
        menuFile.addSeparator()
        menuFile.addAction(self.actionSave)
        menuFile.addAction(self.actionSaveAs)
        menuFile.addSeparator()
        menuFile.addAction(self.actionNextTab)
        menuFile.addAction(self.actionPrevTab)
        menuFile.addAction(self.actionCloseCurrentTab)
        menuFile.addSeparator()
        menuExport = menuFile.addMenu(self.tr('Export'))
        menuExport.addAction(self.actionSaveHtml)
        menuExport.addAction(self.actionOdf)
        menuExport.addAction(self.actionPdf)
        if self.extensionActions:
            menuExport.addSeparator()
            for action, mimetype in self.extensionActions:
                menuExport.addAction(action)
            menuExport.aboutToShow.connect(self.updateExtensionsVisibility)
        menuFile.addAction(self.actionPrint)
        menuFile.addAction(self.actionPrintPreview)
        menuFile.addSeparator()
        menuFile.addAction(self.actionQuit)
        menuEdit.addAction(self.actionUndo)
        menuEdit.addAction(self.actionRedo)
        menuEdit.addSeparator()
        menuEdit.addAction(self.actionCut)
        menuEdit.addAction(self.actionCopy)
        menuEdit.addAction(self.actionPaste)
        menuEdit.addAction(self.actionPasteImage)
        menuEdit.addSeparator()
        menuEdit.addAction(self.actionMoveUp)
        menuEdit.addAction(self.actionMoveDown)
        menuEdit.addSeparator()
        if enchant is not None:
            menuSC = menuEdit.addMenu(self.tr('Spell check'))
            menuSC.addAction(self.actionEnableSC)
            menuSC.addAction(self.actionSetLocale)
        menuEdit.addAction(self.actionSearch)
        menuEdit.addAction(self.actionGoToLine)
        menuEdit.addAction(self.actionChangeEditorFont)
        menuEdit.addAction(self.actionChangePreviewFont)
        menuEdit.addSeparator()
        if len(availableMarkups) > 1:
            self.menuMode = menuEdit.addMenu(self.tr('Default markup'))
            for markupAction in markupActions:
                self.menuMode.addAction(markupAction)
        menuFormat = menuEdit.addMenu(self.tr('Formatting'))
        menuFormat.addAction(self.actionBold)
        menuFormat.addAction(self.actionItalic)
        menuFormat.addAction(self.actionUnderline)
        menuEdit.addAction(self.actionWebEngine)
        menuEdit.addSeparator()
        menuEdit.addAction(self.actionViewHtml)
        menuEdit.addAction(self.actionPreview)
        menuEdit.addAction(self.actionLivePreview)
        menuEdit.addAction(self.actionInsertTable)
        menuEdit.addAction(self.actionTableMode)
        menuEdit.addAction(self.actionInsertImages)
        if ReTextFakeVimHandler:
            menuEdit.addAction(self.actionFakeVimMode)
        menuEdit.addSeparator()
        menuEdit.addAction(self.actionFullScreen)
        menuEdit.addAction(self.actionConfig)
        menuHelp.addAction(self.actionHelp)
        menuHelp.addAction(self.actionWhatsNew)
        menuHelp.addSeparator()
        menuHelp.addAction(self.actionAbout)
        menuHelp.addAction(self.actionAboutQt)
        self.toolBar.addAction(self.actionNew)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionOpen)
        self.toolBar.addAction(self.actionSave)
        self.toolBar.addAction(self.actionPrint)
        self.toolBar.addSeparator()
        previewButton = QToolButton(self.toolBar)
        previewButton.setDefaultAction(self.actionPreview)
        previewButton.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        previewButton.setPopupMode(QToolButton.ToolButtonPopupMode.MenuButtonPopup)
        previewButton.setMenu(menuPreview)
        self.toolBar.addWidget(previewButton)
        self.toolBar.addAction(self.actionFullScreen)
        self.editBar.addAction(self.actionUndo)
        self.editBar.addAction(self.actionRedo)
        self.editBar.addSeparator()
        self.editBar.addAction(self.actionCut)
        self.editBar.addAction(self.actionCopy)
        self.editBar.addAction(self.actionPaste)
        self.editBar.addSeparator()
        self.editBar.addWidget(self.formattingBox)
        self.editBar.addWidget(self.symbolBox)
        self.searchEdit = QLineEdit(self.searchBar)
        self.searchEdit.setPlaceholderText(self.tr('Search'))
        self.searchEdit.returnPressed.connect(self.find)
        self.replaceEdit = QLineEdit(self.searchBar)
        self.replaceEdit.setPlaceholderText(self.tr('Replace with'))
        self.replaceEdit.returnPressed.connect(self.find)
        self.csBox = QCheckBox(self.tr('Case sensitively'), self.searchBar)
        self.searchBar.addWidget(self.searchEdit)
        self.searchBar.addWidget(self.replaceEdit)
        self.searchBar.addSeparator()
        self.searchBar.addWidget(self.csBox)
        self.searchBar.addAction(self.actionFindPrev)
        self.searchBar.addAction(self.actionFind)
        replaceButton = QToolButton(self.searchBar)
        replaceButton.setDefaultAction(self.actionReplace)
        replaceButton.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        replaceButton.setPopupMode(QToolButton.ToolButtonPopupMode.MenuButtonPopup)
        replaceButton.setMenu(menuReplace)
        self.searchBar.addWidget(replaceButton)
        self.searchBar.addAction(self.actionCloseSearch)
        self.searchBar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextBesideIcon)
        self.searchBar.setVisible(False)
        self.autoSaveTimer = QTimer(self)
        self.autoSaveTimer.timeout.connect(self.saveAll)
        if globalSettings.autoSave:
            self.autoSaveTimer.start(60000)
        self.ind = None
        if enchant is not None:
            self.spellCheckLanguages = globalSettings.spellCheckLocale
            languages, errors = self.getSpellCheckDictionaries()
            for error in errors:
                warnings.warn(error, RuntimeWarning)
            if not languages:
                globalSettings.spellCheck = False
            if globalSettings.spellCheck:
                self.actionEnableSC.setChecked(True)
        self.fileSystemWatcher = QFileSystemWatcher()
        self.fileSystemWatcher.fileChanged.connect(self.fileChanged)

    def restoreLastOpenedFiles(self):
        for file in globalCache.lastFileList:
            self.openFileWrapper(file)

        # Show the tab of last opened file
        lastTabIndex = globalCache.lastTabIndex
        if lastTabIndex >= 0 and lastTabIndex < self.tabWidget.count():
            self.tabWidget.setCurrentIndex(lastTabIndex)

    def iterateTabs(self):
        for i in range(self.tabWidget.count()):
            yield self.tabWidget.widget(i)

    def updateStyleSheet(self):
        self.ss = None
        if globalSettings.styleSheet:
            with open(globalSettings.styleSheet) as sheetfile:
                self.ss = sheetfile.read()

    def initTabWidget(self):
        def dragEnterEvent(e):
            e.acceptProposedAction()
        def dropEvent(e):
            fn = bytes(e.mimeData().data('text/plain')).decode().rstrip()
            if fn.startswith('file:'):
                fn = QUrl(fn).toLocalFile()
            self.openFileWrapper(fn)
        self.tabWidget.setTabsClosable(True)
        self.tabWidget.setAcceptDrops(True)
        self.tabWidget.setMovable(True)
        self.tabWidget.dragEnterEvent = dragEnterEvent
        self.tabWidget.dropEvent = dropEvent
        self.tabWidget.setTabBarAutoHide(globalSettings.tabBarAutoHide)

    def initDirectoryTree(self):
        path = globalSettings.directoryPath
        self.fileSystemModel = ReTextFileSystemModel(self.treeView)
        self.fileSystemModel.setRootPath(path)
        supportedExtensions = ['.txt']
        for markup in markups.get_all_markups():
            supportedExtensions += markup.file_extensions
        filters = ["*" + s for s in supportedExtensions]
        self.fileSystemModel.setNameFilters(filters)
        self.fileSystemModel.setNameFilterDisables(False)
        self.treeView.setModel(self.fileSystemModel)
        self.treeView.setRootIndex(self.fileSystemModel.index(path))
        self.treeView.setColumnHidden(1, True)
        self.treeView.setColumnHidden(2, True)
        self.treeView.setColumnHidden(3, True)
        self.treeView.setHeaderHidden(True)

    def treeItemSelected(self, signal):
        file_path = self.fileSystemModel.filePath(signal)
        if os.path.isdir(file_path):
            return
        self.openFileWrapper(file_path)

    def act(self, name, icon=None, trig=None, trigbool=None, shct=None):
        if not isinstance(shct, QKeySequence):
            shct = QKeySequence(shct)
        if icon:
            action = QAction(self.actIcon(icon), name, self)
        else:
            action = QAction(name, self)
        if trig:
            action.triggered.connect(trig)
        elif trigbool:
            action.setCheckable(True)
            action.triggered[bool].connect(trigbool)
        if shct:
            action.setShortcut(shct)
        return action

    def actIcon(self, name):
        return QIcon.fromTheme(name, QIcon(getBundledIcon(name)))

    def printError(self):
        print('Exception occurred while parsing document:', file=sys.stderr)
        traceback.print_exc()

    def updateTabTitle(self, ind, tab):
        changed = tab.editBox.document().isModified()
        if changed:
            title = tab.getBaseName() + '*'
        else:
            title = tab.getBaseName()
        self.tabWidget.setTabText(ind, title)

    def tabFileNameChanged(self, tab):
        '''
        Perform all UI state changes that need to be done when the
        filename of the current tab has changed.
        '''
        if tab == self.currentTab:
            if tab.fileName:
                self.setWindowTitle("")
                if globalSettings.windowTitleFullPath:
                    self.setWindowTitle(tab.fileName + '[*]')
                self.setWindowFilePath(tab.fileName)
                self.updateTabTitle(self.ind, tab)
                self.tabWidget.setTabToolTip(self.ind, tab.fileName)
                QDir.setCurrent(QFileInfo(tab.fileName).dir().path())
            else:
                self.setWindowFilePath('')
                self.setWindowTitle(self.tr('New document') + '[*]')

            canReload = bool(tab.fileName) and not tab.autoSaveActive()
            self.actionSetEncoding.setEnabled(canReload)
            self.actionReload.setEnabled(canReload)

    def tabActiveMarkupChanged(self, tab):
        '''
        Perform all UI state changes that need to be done when the
        active markup class of the current tab has changed.
        '''
        if tab == self.currentTab:
            markupClass = tab.getActiveMarkupClass()
            dtMarkdown = (markupClass == markups.MarkdownMarkup)
            dtMkdOrReST = dtMarkdown or (markupClass == markups.ReStructuredTextMarkup)
            self.formattingBox.setEnabled(dtMarkdown)
            self.symbolBox.setEnabled(dtMarkdown)
            self.actionUnderline.setEnabled(dtMarkdown)
            self.actionBold.setEnabled(dtMkdOrReST)
            self.actionItalic.setEnabled(dtMkdOrReST)

    def tabModificationStateChanged(self, tab):
        '''
        Perform all UI state changes that need to be done when the
        modification state of the current tab has changed.
        '''

        if tab == self.currentTab:
            changed = tab.editBox.document().isModified()
            self.actionSave.setEnabled(changed)
            self.updateTabTitle(self.ind, tab)
            self.setWindowModified(changed)

    def createTab(self, fileName):
        previewState = previewStatesByName.get(globalSettings.defaultPreviewState, PreviewDisabled)
        if previewState == PreviewNormal and not fileName:
            previewState = PreviewDisabled  # Opening empty document in preview mode makes no sense
        self.currentTab = ReTextTab(self, fileName, previewState)
        self.currentTab.fileNameChanged.connect(lambda: self.tabFileNameChanged(self.currentTab))
        self.currentTab.modificationStateChanged.connect(
            lambda: self.tabModificationStateChanged(self.currentTab)
        )
        self.currentTab.activeMarkupChanged.connect(lambda: self.tabActiveMarkupChanged(self.currentTab))
        self.tabWidget.addTab(self.currentTab, self.tr("New document"))
        self.currentTab.updateBoxesVisibility()
        if previewState > 0:
            QTimer.singleShot(500, self.currentTab.triggerPreviewUpdate)

    def closeTab(self, ind):
        if self.maybeSave(ind):
            if self.tabWidget.count() == 1:
                self.createTab("")
            closedTab = self.tabWidget.widget(ind)
            if closedTab.fileName:
                self.fileSystemWatcher.removePath(closedTab.fileName)
            self.tabWidget.removeTab(ind)
            closedTab.deleteLater()

    def changeIndex(self, ind):
        '''
        This function is called when a different tab is selected.
        It changes the state of the window to mirror the current state
        of the newly selected tab. Future changes to this state will be
        done in response to signals emitted by the tab, to which the
        window was subscribed when the tab was created. The window is
        subscribed to all tabs like this, but only the active tab will
        logically generate these signals.
        Aside from the above this function also calls the handlers for
        the other changes that are implied by a tab switch: filename
        change, modification state change and active markup change.
        '''
        self.currentTab = self.tabWidget.currentWidget()
        editBox = self.currentTab.editBox
        previewState = self.currentTab.previewState
        self.actionUndo.setEnabled(editBox.document().isUndoAvailable())
        self.actionRedo.setEnabled(editBox.document().isRedoAvailable())
        self.actionCopy.setEnabled(editBox.textCursor().hasSelection())
        self.actionCut.setEnabled(editBox.textCursor().hasSelection())
        self.actionPreview.setChecked(previewState >= PreviewLive)
        self.actionLivePreview.setChecked(previewState == PreviewLive)
        self.actionTableMode.setChecked(editBox.tableModeEnabled)
        self.editBar.setEnabled(previewState < PreviewNormal)
        self.ind = ind
        editBox.setFocus(Qt.FocusReason.OtherFocusReason)

        self.tabFileNameChanged(self.currentTab)
        self.tabModificationStateChanged(self.currentTab)
        self.tabActiveMarkupChanged(self.currentTab)

    def changeEditorFont(self):
        font, ok = QFontDialog.getFont(globalSettings.getEditorFont(), self)
        if ok:
            self.setEditorFont(font)

    def setEditorFont(self, font):
        globalSettings.editorFont = font.toString()
        for tab in self.iterateTabs():
            tab.editBox.updateFont()

    def changePreviewFont(self):
        font, ok = QFontDialog.getFont(globalSettings.getPreviewFont(), self)
        if ok:
            self.setPreviewFont(font)

    def setPreviewFont(self, font):
        globalSettings.font = font.toString()
        for tab in self.iterateTabs():
            tab.triggerPreviewUpdate()

    def preview(self, viewmode):
        self.currentTab.previewState = viewmode * 2
        self.actionLivePreview.setChecked(False)
        self.editBar.setDisabled(viewmode)
        self.currentTab.updateBoxesVisibility()
        self.currentTab.triggerPreviewUpdate()

    def enableLivePreview(self, livemode):
        self.currentTab.previewState = int(livemode)
        self.actionPreview.setChecked(livemode)
        self.editBar.setEnabled(True)
        self.currentTab.updateBoxesVisibility()
        self.currentTab.triggerPreviewUpdate()

    def enableWebEngine(self, enable):
        globalSettings.useWebEngine = enable
        for tab in self.iterateTabs():
            tab.rebuildPreviewBox()

    def enableCopy(self, copymode):
        self.actionCopy.setEnabled(copymode)
        self.actionCut.setEnabled(copymode)

    def enableFullScreen(self, yes):
        if yes:
            self.showFullScreen()
        else:
            self.showNormal()

    def openConfigDialog(self):
        dlg = ConfigDialog(self)
        dlg.setWindowTitle(self.tr('Preferences'))
        dlg.show()

    def enableFakeVimMode(self, yes):
        globalSettings.useFakeVim = yes
        if yes:
            FakeVimMode.init(self)
            for tab in self.iterateTabs():
                tab.editBox.installFakeVimHandler()
        else:
            FakeVimMode.exit(self)

    def getSpellCheckDictionaries(self):
        """Returns the list of languages and the list of errors."""
        dictionaries = []
        errors = []
        if enchant is None:
            return dictionaries, errors
        for language in self.spellCheckLanguages.split(','):
            try:
                dict = enchant.Dict(language.strip() or None)
            except enchant.errors.Error as e:
                errors.append(str(e))
            else:
                dictionaries.append(dict)
        return dictionaries, errors

    def enableSpellCheck(self, yes):
        dictionaries = []
        if yes:
            dictionaries, errors = self.getSpellCheckDictionaries()
            if errors:
                QMessageBox.warning(self, '', '\n'.join(errors))
        for tab in self.iterateTabs():
            tab.highlighter.dictionaries = dictionaries
            tab.highlighter.rehighlight()
        globalSettings.spellCheck = bool(dictionaries)
        self.actionEnableSC.setChecked(bool(dictionaries))

    def changeLocale(self):
        localedlg = LocaleDialog(self, defaultText=self.spellCheckLanguages)
        if localedlg.exec() != QDialog.DialogCode.Accepted:
            return
        self.spellCheckLanguages = localedlg.localeEdit.text()
        self.enableSpellCheck(True)
        globalSettings.spellCheckLocale = self.spellCheckLanguages

    def search(self):
        self.searchBar.setVisible(True)
        self.searchEdit.setFocus(Qt.FocusReason.ShortcutFocusReason)

    def goToLine(self):
        line, ok = QInputDialog.getInt(self, self.tr("Go to line"), self.tr("Type the line number"))
        if ok:
            self.currentTab.goToLine(line-1)

    def searchBarVisibilityChanged(self, visible):
        if visible:
            self.searchEdit.setFocus(Qt.FocusReason.ShortcutFocusReason)

    def find(self, back=False, replace=False):
        flags = QTextDocument.FindFlag(0)
        if back:
            flags |= QTextDocument.FindFlag.FindBackward
        if self.csBox.isChecked():
            flags |= QTextDocument.FindFlag.FindCaseSensitively
        text = self.searchEdit.text()
        replaceText = self.replaceEdit.text() if replace else None
        found = self.currentTab.find(text, flags, replaceText=replaceText)
        self.setSearchEditColor(found)

    def replaceAll(self):
        text = self.searchEdit.text()
        replaceText = self.replaceEdit.text()
        found = self.currentTab.replaceAll(text, replaceText)
        self.setSearchEditColor(found)

    def setSearchEditColor(self, found):
        palette = self.searchEdit.palette()
        palette.setColor(QPalette.ColorGroup.Active, QPalette.ColorRole.Base,
                         Qt.GlobalColor.white if found else QColor(255, 102, 102))
        self.searchEdit.setPalette(palette)

    def showInDir(self):
        if self.currentTab.fileName:
            path = QFileInfo(self.currentTab.fileName).path()
            QDesktopServices.openUrl(QUrl.fromLocalFile(path))
        else:
            QMessageBox.warning(self, '', self.tr("Please, save the file somewhere."))

    def moveToTopOfRecentFileList(self, fileName):
        if fileName:
            files = globalCache.recentFileList
            if fileName in files:
                files.remove(fileName)
            files.insert(0, fileName)
            recentCount = globalSettings.recentDocumentsCount
            if len(files) > recentCount:
                del files[recentCount:]
            globalCache.recentFileList = files

    def createNew(self, text=None):
        self.createTab("")
        self.ind = self.tabWidget.count()-1
        self.tabWidget.setCurrentIndex(self.ind)
        if text:
            self.currentTab.editBox.textCursor().insertText(text)

    def switchTab(self, shift=1):
        self.tabWidget.setCurrentIndex((self.ind + shift) % self.tabWidget.count())

    def updateRecentFiles(self):
        self.menuRecentFiles.clear()
        self.recentFilesActions = []
        filesOld = globalCache.recentFileList
        files = []
        for f in filesOld:
            if QFile.exists(f):
                files.append(f)
                self.recentFilesActions.append(self.act(f, trig=self.openFunction(f)))
        globalCache.recentFileList = files
        for action in self.recentFilesActions:
            self.menuRecentFiles.addAction(action)

    def markupFunction(self, markup):
        return lambda: self.setDefaultMarkup(markup)

    def openFunction(self, fileName):
        return lambda: self.openFileWrapper(fileName)

    def extensionFunction(self, data):
        return lambda: \
        self.runExtensionCommand(data['Exec'], data['FileFilter'], data['DefaultExtension'])

    def getExportExtensionsList(self):
        extensions = []
        datadirs = QStandardPaths.standardLocations(
            QStandardPaths.StandardLocation.GenericDataLocation)
        for datadir in datadirs:
            extsdir = QDir(os.path.join(datadir, 'retext', 'export-extensions'))
            if extsdir.exists():
                for fileInfo in extsdir.entryInfoList(['*.desktop', '*.ini'],
                QDir.Filter.Files | QDir.Filter.Readable):
                    extensions.append(self.readExtension(fileInfo.filePath()))
        locale = QLocale.system().name()
        self.extensionActions = []
        for extension in extensions:
            try:
                locale_short = locale.split('_')[0]
                if f'Name[{locale}]' in extension:
                    name = extension[f'Name[{locale}]']
                elif f'Name[{locale_short}]' in extension:
                    name = extension[f'Name[{locale_short}]']
                else:
                    name = extension['Name']
                data = {}
                for prop in ('FileFilter', 'DefaultExtension', 'Exec'):
                    if 'X-ReText-'+prop in extension:
                        data[prop] = extension['X-ReText-'+prop]
                    elif prop in extension:
                        data[prop] = extension[prop]
                    else:
                        data[prop] = ''
                action = self.act(name, trig=self.extensionFunction(data))
                if 'Icon' in extension:
                    action.setIcon(self.actIcon(extension['Icon']))
                mimetype = extension['MimeType'] if 'MimeType' in extension else None
            except KeyError:
                print('Failed to parse extension: Name is required', file=sys.stderr)
            else:
                self.extensionActions.append((action, mimetype))

    def updateExtensionsVisibility(self):
        markupClass = self.currentTab.getActiveMarkupClass()
        for action in self.extensionActions:
            if markupClass is None:
                action[0].setEnabled(False)
                continue
            mimetype = action[1]
            if mimetype is None:
                enabled = True
            elif markupClass == markups.MarkdownMarkup:
                enabled = (mimetype == "text/markdown")
            elif markupClass == markups.ReStructuredTextMarkup:
                enabled = (mimetype == "text/x-rst")
            else:
                enabled = False
            action[0].setEnabled(enabled)

    def readExtension(self, fileName):
        parser = configparser.ConfigParser(interpolation=None)
        parser.optionxform = str
        parser.read(fileName)
        return dict(parser.items('Desktop Entry'))

    def openFile(self):
        supportedExtensions = ['.txt']
        for markup in markups.get_all_markups():
            supportedExtensions += markup.file_extensions
        fileFilter = ' (' + ' '.join('*' + ext for ext in supportedExtensions) + ');;'
        fileNames = QFileDialog.getOpenFileNames(self,
            self.tr("Select one or several files to open"), QDir.currentPath(),
            self.tr("Supported files") + fileFilter + self.tr("All files (*)"))
        for fileName in fileNames[0]:
            self.openFileWrapper(fileName)

    @pyqtSlot(str)
    def openFileWrapper(self, fileName):
        if not fileName:
            return
        fileName = QFileInfo(fileName).canonicalFilePath()
        exists = False
        for i, tab in enumerate(self.iterateTabs()):
            if tab.fileName == fileName:
                exists = True
                ex = i
        if exists:
            self.tabWidget.setCurrentIndex(ex)
        elif QFile.exists(fileName):
            noEmptyTab = (
                (self.ind is None) or
                self.currentTab.fileName or
                self.currentTab.editBox.toPlainText() or
                self.currentTab.editBox.document().isModified()
            )
            if noEmptyTab:
                self.createTab(fileName)
                self.ind = self.tabWidget.count()-1
                self.tabWidget.setCurrentIndex(self.ind)
            elif globalSettings.defaultPreviewState == "normal-preview":
                self.actionPreview.setChecked(True)
                self.preview(True)
            if fileName:
                self.fileSystemWatcher.addPath(fileName)
            self.currentTab.readTextFromFile(fileName)
            self.moveToTopOfRecentFileList(self.currentTab.fileName)

    def showEncodingDialog(self):
        if not self.maybeSave(self.ind):
            return
        dialog = EncodingDialog(self)
        if dialog.exec() == QDialog.DialogCode.Accepted:
            encoding = dialog.encodingEdit.text()
            self.currentTab.readTextFromFile(None, encoding)

    def saveFileAs(self):
        self.saveFile(dlg=True)

    def saveAll(self):
        for tab in self.iterateTabs():
            if tab.autoSaveActive() and tab.editBox.document().isModified():
                tab.saveTextToFile()

    def saveFile(self, dlg=False):
        fileNameToSave = self.currentTab.fileName

        if (not fileNameToSave) or dlg:
            proposedFileName = ""
            markupClass = self.currentTab.getActiveMarkupClass()
            if (markupClass is None) or not hasattr(markupClass, 'default_extension'):
                defaultExt = self.tr("Plain text (*.txt)")
                ext = ".txt"
            else:
                defaultExt = (
                    self.tr('%s files', 'Example of final string: Markdown files') % markupClass.name
                    + ' ('
                    + ' '.join('*' + extension for extension in markupClass.file_extensions)
                    + ')'
                )
                if markupClass == markups.MarkdownMarkup:
                    ext = globalSettings.markdownDefaultFileExtension
                elif markupClass == markups.ReStructuredTextMarkup:
                    ext = globalSettings.restDefaultFileExtension
                else:
                    ext = markupClass.default_extension
            if fileNameToSave is not None:
                proposedFileName = fileNameToSave
            fileNameToSave = QFileDialog.getSaveFileName(self,
                self.tr("Save file"), proposedFileName, defaultExt)[0]
            if fileNameToSave:
                if not QFileInfo(fileNameToSave).suffix():
                    fileNameToSave += ext
                # Make sure we don't overwrite a file opened in other tab
                for tab in self.iterateTabs():
                    if tab is not self.currentTab and tab.fileName == fileNameToSave:
                        QMessageBox.warning(self, "",
                            self.tr("Cannot save to file which is open in another tab!"))
                        return False
                self.actionSetEncoding.setDisabled(self.currentTab.autoSaveActive())
        if fileNameToSave:
            if self.currentTab.saveTextToFile(fileNameToSave):
                self.moveToTopOfRecentFileList(self.currentTab.fileName)
                return True
        return False

    def saveHtml(self, fileName):
        if not QFileInfo(fileName).suffix():
            fileName += ".html"
        try:
            _, htmltext, _ = self.currentTab.getDocumentForExport(webenv=True)
        except Exception:
            return self.printError()

        encoding = globalSettings.defaultCodec or None
        try:
            with open(fileName, 'w', encoding=encoding) as htmlFile:
                htmlFile.write(htmltext)
        except (OSError, UnicodeEncodeError, LookupError) as ex:
            QMessageBox.warning(self, '', str(ex))

    def textDocument(self, title, htmltext):
        td = QTextDocument()
        td.setMetaInformation(QTextDocument.MetaInformation.DocumentTitle, title)
        td.setHtml(htmltext)
        td.setDefaultFont(globalSettings.getPreviewFont())
        return td

    def saveOdf(self):
        title, htmltext, _ = self.currentTab.getDocumentForExport()
        try:
            document = self.textDocument(title, htmltext)
        except Exception:
            return self.printError()
        fileName = QFileDialog.getSaveFileName(self,
            self.tr("Export document to ODT"), self.currentTab.getBaseName() + ".odt",
            self.tr("OpenDocument text files (*.odt)"))[0]
        if not QFileInfo(fileName).suffix():
            fileName += ".odt"
        writer = QTextDocumentWriter(fileName)
        writer.setFormat(b"odf")
        writer.write(document)

    def saveFileHtml(self):
        fileName = QFileDialog.getSaveFileName(self,
            self.tr("Save file"), self.currentTab.getBaseName() + ".html",
            self.tr("HTML files (*.html *.htm)"))[0]
        if fileName:
            self.saveHtml(fileName)

    def getDocumentForPrint(self, title, htmltext, preview):
        try:
            return self.textDocument(title, htmltext)
        except Exception:
            self.printError()

    def standardPrinter(self, title):
        printer = QPrinter(QPrinter.PrinterMode.HighResolution)
        printer.setDocName(title)
        printer.setCreator(f'ReText {app_version}')
        if globalSettings.paperSize:
            pageSize = self.getPageSizeByName(globalSettings.paperSize)
            if pageSize is not None:
                printer.setPageSize(pageSize)
            else:
                QMessageBox.warning(self, '',
                    self.tr('Unrecognized paperSize setting "%s".') %
                    globalSettings.paperSize)
        return printer

    def getPageSizeByName(self, pageSizeName):
        """ Returns a validated QPageSize instance corresponding to the given
        name. Returns None if the name is not a valid PageSize.
        """
        sizesByName = {e.name.lower(): e for e in QPageSize.PageSizeId}
        sizeId = sizesByName.get(pageSizeName.lower())
        if sizeId is not None:
            return QPageSize(sizeId)

    def savePdf(self):
        fileName = QFileDialog.getSaveFileName(self,
            self.tr("Export document to PDF"),
            self.currentTab.getBaseName() + ".pdf",
            self.tr("PDF files (*.pdf)"))[0]
        if fileName:
            if not QFileInfo(fileName).suffix():
                fileName += ".pdf"
            title, htmltext, preview = self.currentTab.getDocumentForExport()
            if globalSettings.useWebEngine:
                pageSize = self.getPageSizeByName(globalSettings.paperSize)
                if pageSize is None:
                    pageSize = QPageSize(QPageSize.PageSizeId.A4)
                margins = QMarginsF(20, 20, 13, 20)  # left, top, right, bottom (in millimeters)
                layout = QPageLayout(
                    pageSize,
                    QPageLayout.Orientation.Portrait,
                    margins,
                    QPageLayout.Unit.Millimeter,
                )
                preview.page().printToPdf(fileName, layout)
                return
            printer = self.standardPrinter(title)
            printer.setOutputFormat(QPrinter.OutputFormat.PdfFormat)
            printer.setOutputFileName(fileName)
            document = self.getDocumentForPrint(title, htmltext, preview)
            if document is not None:
                document.print(printer)

    def printFile(self):
        title, htmltext, preview = self.currentTab.getDocumentForExport()
        printer = self.standardPrinter(title)
        dlg = QPrintDialog(printer, self)
        dlg.setWindowTitle(self.tr("Print document"))
        if (dlg.exec() == QDialog.DialogCode.Accepted):
            document = self.getDocumentForPrint(title, htmltext, preview)
            if document is not None:
                document.print(printer)

    def printPreview(self):
        title, htmltext, preview = self.currentTab.getDocumentForExport()
        document = self.getDocumentForPrint(title, htmltext, preview)
        if document is None:
            return
        printer = self.standardPrinter(title)
        preview = QPrintPreviewDialog(printer, self)
        preview.paintRequested.connect(document.print)
        preview.exec()

    def runExtensionCommand(self, command, filefilter, defaultext):
        of = ('%of' in command)
        html = ('%html' in command)
        if of:
            if defaultext and not filefilter:
                filefilter = '*'+defaultext
            fileName = QFileDialog.getSaveFileName(self,
                self.tr('Export document'), '', filefilter)[0]
            if not fileName:
                return
            if defaultext and not QFileInfo(fileName).suffix():
                fileName += defaultext
        else:
            fileName = 'out' + defaultext
        basename = f'.{self.currentTab.getBaseName()}.retext-temp'
        if html:
            tmpname = basename+'.html'
            self.saveHtml(tmpname)
        else:
            tmpname = basename + self.currentTab.getActiveMarkupClass().default_extension
            self.currentTab.writeTextToFile(tmpname)
        command = command.replace('%of', shlex.quote(fileName))
        command = command.replace('%html' if html else '%if', shlex.quote(tmpname))
        try:
            Popen(str(command), shell=True).wait()
        except Exception as error:
            errorstr = str(error)
            QMessageBox.warning(self, '', self.tr('Failed to execute the command:')
            + '\n' + errorstr)
        QFile(tmpname).remove()

    def clipboardDataChanged(self):
        mimeData = QApplication.instance().clipboard().mimeData()
        if mimeData is not None:
            self.actionPaste.setEnabled(mimeData.hasText())
            self.actionPasteImage.setEnabled(mimeData.hasImage())

    def insertFormatting(self, formatting):
        if formatting == 'table':
            dialog = InsertTableDialog(self)
            dialog.show()
            self.formattingBox.setCurrentIndex(0)
            return

        cursor = self.currentTab.editBox.textCursor()
        text = cursor.selectedText()
        moveCursorTo = None

        def c(cursor):
            nonlocal moveCursorTo
            moveCursorTo = cursor.position()

        def ensurenl(cursor):
            if not cursor.atBlockStart():
                cursor.insertText('\n\n')

        toinsert = {
            'header': (ensurenl, '# ', text),
            'italic': ('*', text, c, '*'),
            'bold': ('**', text, c, '**'),
            'underline': ('<u>', text, c, '</u>'),
            'numbering': (ensurenl, ' 1. ', text),
            'bullets': (ensurenl, '  * ', text),
            'image': ('![', text or self.tr('Alt text'), c, '](', self.tr('URL'), ')'),
            'link': ('[', text or self.tr('Link text'), c, '](', self.tr('URL'), ')'),
            'inline code': ('`', text, c, '`'),
            'code block': (ensurenl, '    ', text),
            'blockquote': (ensurenl, '> ', text),
        }

        if formatting not in toinsert:
            return

        cursor.beginEditBlock()
        for token in toinsert[formatting]:
            if callable(token):
                token(cursor)
            else:
                cursor.insertText(token)
        cursor.endEditBlock()

        self.formattingBox.setCurrentIndex(0)
        # Bring back the focus on the editor
        self.currentTab.editBox.setFocus(Qt.FocusReason.OtherFocusReason)

        if moveCursorTo:
            cursor.setPosition(moveCursorTo)
            self.currentTab.editBox.setTextCursor(cursor)

    def insertSymbol(self, num):
        if num:
            self.currentTab.editBox.insertPlainText('&'+self.usefulChars[num-1]+';')
        self.symbolBox.setCurrentIndex(0)

    def fileChanged(self, fileName):
        tab = None
        for testtab in self.iterateTabs():
            if testtab.fileName == fileName:
                tab = testtab
        if tab is None:
            self.fileSystemWatcher.removePath(fileName)
            return
        if not QFile.exists(fileName):
            self.tabWidget.setCurrentWidget(tab)
            tab.editBox.document().setModified(True)
            QMessageBox.warning(self, '', self.tr(
                'This file has been deleted by other application.\n'
                'Please make sure you save the file before exit.'))
        elif not tab.editBox.document().isModified():
            # File was not modified in ReText, reload silently
            tab.readTextFromFile()
        else:
            tab.forceDisableAutoSave = True
            self.tabWidget.setCurrentWidget(tab)
            text = self.tr(
                'This document has been modified by other application.\n'
                'Do you want to reload the file (this will discard all '
                'your changes)?\n')
            if globalSettings.autoSave:
                text += self.tr(
                    'Automatic saving has been temporarily disabled for this '
                    'tab to prevent data loss. It will be re-enabled when you '
                    'reload the file or save it manually.')
            messageBox = QMessageBox(QMessageBox.Icon.Warning, '', text)
            reloadButton = messageBox.addButton(self.tr('Reload'), QMessageBox.ButtonRole.YesRole)
            messageBox.addButton(QMessageBox.StandardButton.Cancel)
            messageBox.exec()
            if messageBox.clickedButton() is reloadButton:
                tab.readTextFromFile()
        if fileName not in self.fileSystemWatcher.files():
            # https://github.com/retext-project/retext/issues/137
            self.fileSystemWatcher.addPath(fileName)

    def maybeSave(self, ind):
        tab = self.tabWidget.widget(ind)
        if tab.autoSaveActive():
            tab.saveTextToFile()
            return True
        if not tab.editBox.document().isModified():
            return True
        self.tabWidget.setCurrentIndex(ind)
        ret = QMessageBox.warning(
            self,
            '',
            self.tr("The document has been modified.\nDo you want to save your changes?"),
            QMessageBox.StandardButton.Save
            | QMessageBox.StandardButton.Discard
            | QMessageBox.StandardButton.Cancel,
        )
        if ret == QMessageBox.StandardButton.Save:
            return self.saveFile(False)
        elif ret == QMessageBox.StandardButton.Cancel:
            return False
        return True

    def closeEvent(self, closeevent):
        for ind in range(self.tabWidget.count()):
            if not self.maybeSave(ind):
                return closeevent.ignore()
        if globalSettings.saveWindowGeometry:
            globalCache.windowGeometry = self.saveGeometry()
            if self.treeView.isVisible():
                globalCache.splitterState = self.splitter.saveState()
            else:
                globalCache.splitterState = QByteArray()
        else:
            globalCache.windowGeometry = QByteArray()
            globalCache.splitterState = QByteArray()
        if globalSettings.openLastFilesOnStartup:
            files = [tab.fileName for tab in self.iterateTabs()]
            globalCache.lastFileList = files
            globalCache.lastTabIndex = self.tabWidget.currentIndex()
        closeevent.accept()

    def viewHtml(self):
        htmlDlg = HtmlDialog(self)
        try:
            _, htmltext, _ = self.currentTab.getDocumentForExport(includeStyleSheet=False)
        except Exception:
            return self.printError()
        winTitle = self.currentTab.getBaseName()
        htmlDlg.setWindowTitle(winTitle+" ("+self.tr("HTML code")+")")
        htmlDlg.textEdit.setPlainText(htmltext.rstrip())
        htmlDlg.hl.rehighlight()
        htmlDlg.show()
        htmlDlg.raise_()
        htmlDlg.activateWindow()

    def insertImages(self):
        supportedExtensions = ['.png', '.jpg', '.jpeg', '.gif', '.bmp']
        fileFilter = ' (' + ' '.join('*' + ext for ext in supportedExtensions) + ');;'
        fileNames, _selectedFilter = QFileDialog.getOpenFileNames(self,
            self.tr("Select one or several images to open"), QDir.currentPath(),
            self.tr("Supported files") + fileFilter + self.tr("All files (*)"))

        cursor = self.currentTab.editBox.textCursor()

        imagesMarkup = '\n'.join(
            self.currentTab.editBox.getImageMarkup(fileName)
            for fileName in fileNames)
        cursor.insertText(imagesMarkup)

        self.formattingBox.setCurrentIndex(0)
        self.currentTab.editBox.setFocus(Qt.FocusReason.OtherFocusReason)

    def openHelp(self):
        QDesktopServices.openUrl(QUrl('https://github.com/retext-project/retext/wiki'))

    def openReleases(self):
        QDesktopServices.openUrl(QUrl('https://github.com/retext-project/retext/releases'))

    def aboutDialog(self):
        QMessageBox.about(self, self.aboutWindowTitle,
        '<p><b>' + (self.tr('ReText %s (using PyMarkups %s)') % (app_version, markups.__version__))
        +'</b></p>' + self.tr('Simple but powerful editor'
        ' for Markdown and reStructuredText')
        +'</p><p>'+self.tr('Author: Dmitry Shachnev, 2011').replace('2011', '2011–2025')
        +'<br><a href="https://github.com/retext-project/retext">GitHub</a> | '
        +'<a href="https://daringfireball.net/projects/markdown/syntax">'
        +self.tr('Markdown syntax')
        +'</a> | <a href="https://docutils.sourceforge.io/docs/user/rst/quickref.html">'
        +self.tr('reStructuredText syntax')+'</a></p>')

    def setDefaultMarkup(self, markupClass):
        globalSettings.defaultMarkup = markupClass.name
        for tab in self.iterateTabs():
            if not tab.fileName:
                tab.updateActiveMarkupClass()



# This file is part of ReText
# Copyright: 2016 Maurice van der Pot
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import multiprocessing as mp
import os
import pickle
import signal
import struct
import traceback
import weakref
from socket import socketpair

import markups
from PyQt6.QtCore import QObject, QSocketNotifier, pyqtSignal


def recvall(sock, remaining):
    alldata = bytearray()
    while remaining > 0:
        data = sock.recv(remaining)
        if len(data) == 0:
            raise EOFError(
                'Received 0 bytes from socket while more bytes were expected.'
                ' Did the sender process exit unexpectedly?'
            )
        alldata.extend(data)
        remaining -= len(data)

    return alldata

def receiveObject(sock):
    sizeBuf = recvall(sock, 4)
    size = struct.unpack('I', sizeBuf)[0]
    message = recvall(sock, size)
    obj = pickle.loads(message)
    return obj

def sendObject(sock, obj):
    message = pickle.dumps(obj)
    sizeBuf = struct.pack('I', len(message))
    sock.sendall(sizeBuf)
    sock.sendall(message)

class ConversionError(Exception):
    pass

class MarkupNotAvailableError(Exception):
    pass

def _indent(text, prefix):
    return ''.join(f'{prefix}{line}\n' for line in text.splitlines())

def _converter_process_func(conn_parent, conn_child):
    conn_parent.close()

    # Ignore ctrl-C. The main application will also receive the signal and
    # determine if the application should be stopped or not.
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    current_markup = None

    while True:
        job = receiveObject(conn_child)
        if job['command'] == 'quit':
            break
        elif job['command'] == 'convert':
            try:
                os.chdir(job['current_dir'])
                if (not current_markup or
                    current_markup.name != job['markup_name'] or
                    current_markup.filename != job['filename']):
                    markup_class = markups.find_markup_class_by_name(job['markup_name'])
                    if not markup_class.available():
                        raise MarkupNotAvailableError('The specified markup was not available')

                    current_markup = markup_class(job['filename'])
                    current_markup.requested_extensions = job['requested_extensions']

                converted = current_markup.convert(job['text'])
                result = ('ok', converted)
            except MarkupNotAvailableError as e:
                result = ('markupnotavailableerror', e.args)
            except Exception:
                result = ('conversionerror',
                          'The background markup conversion process received this exception:\n' +
                          _indent(traceback.format_exc(), '    '))

            try:
                sendObject(conn_child, result)
            except BrokenPipeError:
                # Continue despite the broken pipe because we expect that a
                # 'quit' command will have been sent. If it has been then we
                # should terminate without any error messages. If no command
                # was queued we will get an EOFError from the read, giving us a
                # second chance to show that something went wrong by exiting
                # with a traceback.
                continue


class ConverterProcess(QObject):

    conversionDone = pyqtSignal()

    def __init__(self):
        super().__init__()

        conn_parent, conn_child = socketpair()

        # TODO: figure out which of the two sockets should be set to
        #       inheritable and which should be passed to the child
        conn_child.set_inheritable(True)

        # Use a local variable for child so that we can talk to the child in
        # on_finalize without needing a reference to self
        child = mp.Process(target=_converter_process_func, args=(conn_parent, conn_child))
        child.daemon = True
        child.start()
        self.child = child

        conn_child.close()
        self.conn = conn_parent

        self.busy = False
        self.notificationPending = False
        self.conversionNotifier = QSocketNotifier(self.conn.fileno(),
                                                  QSocketNotifier.Type.Read)
        self.conversionNotifier.activated.connect(self._conversionNotifierActivated)

        def on_finalize(conn):
            sendObject(conn_parent, {'command':'quit'})
            conn_parent.close()
            child.join()

        weakref.finalize(self, on_finalize, conn_parent)

    def _conversionNotifierActivated(self):
        # The ready-for-read signal on the socket may be triggered multiple
        # times, but we only send a single notification to the client as soon
        # as the results of the conversion are starting to come in. This makes
        # it easy for clients to avoid multiple calls to get_result for the
        # same conversion.
        if self.notificationPending:
            self.notificationPending = False

            # Set the socket to blocking before waking up any interested parties,
            # because it has been set to unblocking by QSocketNotifier
            self.conn.setblocking(True)
            self.conversionDone.emit()

    def start_conversion(self, markup_name, filename, requested_extensions, text, current_dir):
        if self.busy:
            raise RuntimeError('Already converting')

        sendObject(self.conn, {'command': 'convert',
                               'markup_name' : markup_name,
                               'filename' : filename,
                               'current_dir': current_dir,
                               'requested_extensions' : requested_extensions,
                               'text' : text})
        self.busy = True
        self.notificationPending = True

    def get_result(self):
        if not self.busy:
            raise RuntimeError('No ongoing conversion')

        self.busy = False

        status, result = receiveObject(self.conn)

        if status == 'markupnotavailableerror':
            raise MarkupNotAvailableError(result)
        elif status == 'conversionerror':
            raise ConversionError(result)

        return result

    def stop(self):
        sendObject(self.conn, {'command': 'quit'})
        self.conn.close()




# vim: ts=4:sw=4:expandtab

# This file is part of ReText
# Copyright: 2012-2025 Dmitry Shachnev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from os.path import abspath, dirname, expanduser, join

import markups
import markups.common
from PyQt6.QtCore import QByteArray, QLocale, QSettings
from PyQt6.QtGui import QFont, QFontDatabase

app_version = "8.1.0"

settings = QSettings('ReText project', 'ReText')

if not str(settings.fileName()).endswith('.conf'):
    # We are on Windows probably
    settings = QSettings(QSettings.Format.IniFormat, QSettings.Scope.UserScope,
        'ReText project', 'ReText')

cache = QSettings('ReText project', 'cache')

if not str(cache.fileName()).endswith('.conf'):
    # We are on Windows probably
    cache = QSettings(QSettings.Format.IniFormat, QSettings.Scope.UserScope,
        'ReText project', 'cache')


packageDir = abspath(dirname(__file__))

def getBundledIcon(iconName):
    return join(packageDir, 'icons', iconName + '.png')


configOptions = {
    'appStyleSheet': '',
    'autoSave': False,
    'defaultCodec': '',
    'defaultMarkup': markups.MarkdownMarkup.name,
    'defaultPreviewState': 'editor',
    'detectEncoding': True,
    'directoryPath': expanduser("~"),
    'documentStatsEnabled': False,
    'editorFont': '',
    'font': '',
    'handleWebLinks': False,
    'hideToolBar': False,
    'highlightCurrentLine': 'disabled',
    'iconTheme': '',
    'lineNumbersEnabled': False,
    'markdownDefaultFileExtension': '.mkd',
    'openFilesInExistingWindow': True,
    'openLastFilesOnStartup': False,
    'orderedListMode': 'increment',
    'paperSize': '',
    'pygmentsStyle': 'default',
    'recentDocumentsCount': 10,
    'relativeLineNumbers': False,
    'restDefaultFileExtension': '.rst',
    'rightMargin': 0,
    'rightMarginWrap': False,
    'saveWindowGeometry': False,
    'showDirectoryTree': False,
    'spellCheck': False,
    'spellCheckLocale': '',
    'styleSheet': '',
    'syncScroll': True,
    'tabBarAutoHide': False,
    'tabInsertsSpaces': True,
    'tabWidth': 4,
    'uiLanguage': QLocale.system().name(),
    'useFakeVim': False,
    'useWebEngine': False,
    'wideCursor': False,
    'windowTitleFullPath': False,
}

cacheOptions = {
    'lastFileList': [],
    'lastTabIndex': 0,
    'recentFileList': [],
    'splitterState': QByteArray(),
    'webEngineZoomFactor': 1.0,
    'windowGeometry': QByteArray(),
}

def readFromSettings(key, keytype, settings, default=None):
    if not settings.contains(key):
        return default
    try:
        value = settings.value(key, type=keytype)
        if isinstance(value, keytype):
            return value
        return keytype(value)
    except TypeError as error:
        # Type mismatch
        print('Warning: '+str(error))
        # Return an instance of keytype
        return default if (default is not None) else keytype()

def readListFromSettings(key, settings):
    if not settings.contains(key):
        return []
    value = settings.value(key)
    if isinstance(value, str):
        return [value]
    else:
        return value

def writeToSettings(key, value, default, settings):
    if value == default:
        settings.remove(key)
    else:
        settings.setValue(key, value)

def writeListToSettings(key, value, settings):
    if len(value) > 1:
        settings.setValue(key, value)
    elif len(value) == 1:
        settings.setValue(key, value[0])
    else:
        settings.remove(key)

def getSettingsFilePath():
    return settings.fileName()


class ReTextSettings:

    def __init__(self, settings, defaults):
        # We have to do this to go around the custom __setattr__ method
        object.__setattr__(self, "settings", settings)
        object.__setattr__(self, "defaults", defaults)
        for option in defaults:
            default = defaults[option]
            if isinstance(default, list):
                object.__setattr__(self, option, readListFromSettings(
                    option, settings=settings))
            else:
                object.__setattr__(self, option, readFromSettings(
                    option, type(default), default=default, settings=settings))

    def __setattr__(self, option, value):
        if option not in self.defaults:
            raise AttributeError('Unknown attribute')
        default = self.defaults[option]
        if isinstance(default, list):
            object.__setattr__(self, option, value.copy())
            writeListToSettings(option, value, settings=self.settings)
        else:
            object.__setattr__(self, option, value)
            writeToSettings(option, value, default=default, settings=self.settings)

    def getPreviewFont(self):
        font = QFont()
        if self.font:
            font.fromString(self.font)
        return font

    def getEditorFont(self):
        font = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)
        if self.editorFont:
            font.fromString(self.editorFont)
        return font

def moveSettingsToCache():
    # Moves the non-editable config options to the cache file
    # This is here for backwards compatibility
    for option in cacheOptions:
        if not cache.contains(option):
            default = cacheOptions[option]
            if isinstance(default, list):
                value = readListFromSettings(option, settings)
                writeListToSettings(option, value, cache)
            else:
                value = readFromSettings(option, type(default), settings, default)
                writeToSettings(option, value, default, cache)
        settings.remove(option)

moveSettingsToCache()
globalSettings = ReTextSettings(settings, configOptions)
globalCache = ReTextSettings(cache, cacheOptions)

markups.common.PYGMENTS_STYLE = globalSettings.pygmentsStyle



# This file is part of ReText
# Copyright: 2018 Changhee Kim, 2018-2024 Dmitry Shachnev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from markups import ReStructuredTextMarkup
from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QDialog, QDialogButtonBox, QGridLayout, QLabel, QSpinBox


class InsertTableDialog(QDialog):
    def __init__(self, parent):
        QDialog.__init__(self, parent)
        self.parent = parent
        self.setWindowTitle(self.tr('Insert table'))
        buttonBox = QDialogButtonBox(self)
        buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Ok |
                                     QDialogButtonBox.StandardButton.Cancel)
        buttonBox.accepted.connect(self.makeTable)
        buttonBox.rejected.connect(self.close)

        layout = QGridLayout(self)

        rowsLabel = QLabel(self.tr('Number of rows') + ':', self)
        columnsLabel = QLabel(self.tr('Number of columns') + ':', self)
        self.rowsSpinBox = QSpinBox(self)
        self.columnsSpinBox = QSpinBox(self)

        self.rowsSpinBox.setRange(1, 10)
        self.columnsSpinBox.setRange(1, 10)
        self.rowsSpinBox.setValue(3)
        self.columnsSpinBox.setValue(3)

        layout.addWidget(rowsLabel, 0, 0)
        layout.addWidget(self.rowsSpinBox, 0, 1, Qt.AlignmentFlag.AlignRight)
        layout.addWidget(columnsLabel, 1, 0)
        layout.addWidget(self.columnsSpinBox, 1, 1, Qt.AlignmentFlag.AlignRight)
        layout.addWidget(buttonBox, 2, 0, 1, 2)

    def makeTable(self):
        rowsCount = self.rowsSpinBox.value()
        columnsCount = self.columnsSpinBox.value() + 1

        tab = self.parent.currentTab
        cursor = tab.editBox.textCursor()

        tableCode = '' if cursor.atBlockStart() else '\n\n'
        if tab.activeMarkupClass == ReStructuredTextMarkup:
            # Insert reStructuredText grid table
            tableCode += '-----'.join('+' * columnsCount) + '\n'
            tableCode += '     '.join('|' * columnsCount) + '\n'
            tableCode += '====='.join('+' * columnsCount) + '\n'
            tableCode += ('     '.join('|' * columnsCount) + '\n' +
                          '-----'.join('+' * columnsCount) + '\n') * rowsCount
        else:
            # Insert Markdown table
            tableCode += '     '.join('|' * columnsCount) + '\n'
            tableCode += '-----'.join('|' * columnsCount) + '\n'
            tableCode += ('     '.join('|' * columnsCount) + '\n') * rowsCount

        cursor.insertText(tableCode)
        self.close()

        # Activate the Table editing mode
        self.parent.actionTableMode.setChecked(True)
        tab.editBox.tableModeEnabled = True



# vim: ts=4:sw=4:expandtab
#
# This file is part of ReText
# Copyright: 2017-2024 Dmitry Shachnev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
from os.path import exists

from PyQt6.QtCore import QDir, QUrl
from PyQt6.QtGui import QDesktopServices, QTextCursor, QTextDocument
from PyQt6.QtWidgets import QTextBrowser

from ReText import globalSettings


class ReTextPreview(QTextBrowser):

    def __init__(self, tab):
        QTextBrowser.__init__(self)
        self.tab = tab
        # if set to True, links to other files will unsuccessfully be opened as anchors
        self.setOpenLinks(False)
        self.anchorClicked.connect(self.openInternal)
        self.lastRenderTime = 0
        self.distToBottom = None
        self.verticalScrollBar().rangeChanged.connect(self.updateScrollPosition)

    def disconnectExternalSignals(self):
        pass

    def openInternal(self, link):
        url = link.url()
        if url.startswith('#'):
            self.scrollToAnchor(url[1:])
            return
        elif link.isRelative():
            fileToOpen = QDir.current().filePath(url)
        else:
            fileToOpen = link.toLocalFile() if link.isLocalFile() else None
        if fileToOpen is not None:
            if exists(fileToOpen):
                link = QUrl.fromLocalFile(fileToOpen)
                if globalSettings.handleWebLinks and fileToOpen.endswith('.html'):
                    self.setSource(link)
                    return
            # This is outside the "if exists" block because we can prompt for
            # creating the file
            if self.tab.openSourceFile(fileToOpen):
                return
        QDesktopServices.openUrl(link)

    def findText(self, text, flags, wrap=False):
        cursor = self.textCursor()
        if wrap and flags & QTextDocument.FindFlag.FindBackward:
            cursor.movePosition(QTextCursor.MoveOperation.End)
        elif wrap:
            cursor.movePosition(QTextCursor.MoveOperation.Start)
        newCursor = self.document().find(text, cursor, flags)
        if not newCursor.isNull():
            self.setTextCursor(newCursor)
            return True
        if not wrap:
            return self.findText(text, flags, wrap=True)
        return False

    def updateScrollPosition(self, minimum, maximum):
        """Called when vertical scroll bar range changes.

        If this happened during preview rendering (less than 0.5s since it
        was started), set the position such that distance to bottom is the
        same as before refresh.
        """
        timeSinceRender = time.time() - self.lastRenderTime
        if timeSinceRender < 0.5 and self.distToBottom is not None and maximum:
            newValue = maximum - self.distToBottom
            if newValue >= minimum:
                self.verticalScrollBar().setValue(newValue)

    def setFont(self, font):
        self.document().setDefaultFont(font)
